"""ChatKit server wiring and cleanup helpers."""

from __future__ import annotations
import asyncio
import logging
import os
from contextlib import suppress
from datetime import UTC, datetime, timedelta
from fastapi import HTTPException, status
from orcheo_backend.app.chatkit import (
    OrcheoChatKitServer,
    create_chatkit_server,
)
from orcheo_backend.app.chatkit_store_postgres import PostgresChatKitStore
from orcheo_backend.app.chatkit_store_sqlite import SqliteChatKitStore
from orcheo_backend.app.chatkit_tokens import (
    ChatKitSessionTokenIssuer,
    ChatKitTokenConfigurationError,
    get_chatkit_token_issuer,
)
from orcheo_backend.app.dependencies import get_repository, get_vault


logger = logging.getLogger(__name__)

_CHATKIT_CLEANUP_INTERVAL_SECONDS = 6 * 60 * 60
_chatkit_cleanup_task: dict[str, asyncio.Task | None] = {"task": None}
_chatkit_server_ref: dict[str, OrcheoChatKitServer | None] = {"server": None}


def _coerce_int(value: object, default: int) -> int:
    if isinstance(value, int):
        return value
    if value is None:
        return default
    try:
        return int(str(value))
    except (TypeError, ValueError):  # pragma: no cover - defensive
        return default


def chatkit_retention_days() -> int:
    from orcheo_backend.app import get_settings

    settings = get_settings()
    days = _coerce_int(settings.get("CHATKIT_RETENTION_DAYS", 30), 30)
    return days if days > 0 else 30


def get_chatkit_store() -> SqliteChatKitStore | PostgresChatKitStore | None:
    server = _chatkit_server_ref.get("server")
    if server is None:
        return None
    store = getattr(server, "store", None)
    if isinstance(store, SqliteChatKitStore | PostgresChatKitStore):
        return store
    return None


async def ensure_chatkit_cleanup_task() -> None:
    """Start the ChatKit cleanup background task if a store is configured."""
    if _chatkit_cleanup_task["task"] is not None:
        return

    from orcheo_backend.app import (
        _CHATKIT_CLEANUP_INTERVAL_SECONDS,
        _get_chatkit_store,
    )
    from orcheo_backend.app import (
        _chatkit_retention_days as retention_days_func,
    )

    store = _get_chatkit_store()
    if store is None:
        return

    retention_days = retention_days_func()
    interval_seconds = max(_CHATKIT_CLEANUP_INTERVAL_SECONDS, 300)

    async def _cleanup_loop() -> None:
        try:
            while True:
                cutoff = datetime.now(tz=UTC) - timedelta(days=retention_days)
                try:
                    pruned = await store.prune_threads_older_than(cutoff)
                    if pruned:
                        logger.info(
                            "Pruned %s ChatKit thread(s) older than %s",
                            pruned,
                            cutoff.isoformat(),
                        )
                except asyncio.CancelledError:
                    raise
                except Exception:  # pragma: no cover
                    logger.exception("ChatKit cleanup task failed")

                await asyncio.sleep(interval_seconds)
        finally:
            _chatkit_cleanup_task["task"] = None

    _chatkit_cleanup_task["task"] = asyncio.create_task(
        _cleanup_loop(),
        name="chatkit_cleanup",
    )


async def cancel_chatkit_cleanup_task() -> None:
    """Cancel the ChatKit cleanup task if it is running."""
    task = _chatkit_cleanup_task.get("task")
    if task is None:
        return

    task.cancel()
    with suppress(asyncio.CancelledError):
        await task
    _chatkit_cleanup_task["task"] = None


def resolve_chatkit_token_issuer() -> ChatKitSessionTokenIssuer:
    """Return the configured ChatKit token issuer or raise a 503 error."""
    try:
        return get_chatkit_token_issuer()
    except ChatKitTokenConfigurationError as exc:
        detail = {
            "message": "ChatKit session token signing key is not configured",
            "code": "chatkit.signing_key_missing",
        }
        raise HTTPException(
            status.HTTP_503_SERVICE_UNAVAILABLE,
            detail,
        ) from exc


def get_chatkit_server() -> OrcheoChatKitServer:
    """Return the singleton ChatKit server wired to the repository."""
    server = _chatkit_server_ref["server"]
    if server is None:
        repository = get_repository()
        server = create_chatkit_server(repository, get_vault)
        _chatkit_server_ref["server"] = server
    return server


def sensitive_logging_enabled() -> bool:
    """Return True when sensitive debug logging should be enabled."""
    dev_environments = {"development", "dev", "local"}
    current_env = os.getenv("ORCHEO_ENV") or os.getenv("NODE_ENV", "production")
    if (current_env or "").lower() in dev_environments:
        return True
    return os.getenv("LOG_SENSITIVE_DEBUG") == "1"


__all__ = [
    "cancel_chatkit_cleanup_task",
    "ensure_chatkit_cleanup_task",
    "get_chatkit_server",
    "resolve_chatkit_token_issuer",
    "sensitive_logging_enabled",
]



"""System metadata routes."""

from __future__ import annotations
from fastapi import APIRouter
from orcheo_backend.app.plugin_inventory import list_runtime_plugins
from orcheo_backend.app.schemas.system import (
    SystemInfoResponse,
    SystemPluginsResponse,
)
from orcheo_backend.app.versioning import get_system_info_payload


public_router = APIRouter()
router = APIRouter()


@public_router.get("/system/health")
def get_system_health() -> dict[str, str]:
    """Return a lightweight unauthenticated health status."""
    return {"status": "ok"}


@router.get("/system/info", response_model=SystemInfoResponse)
def get_system_info() -> SystemInfoResponse:
    """Return current and latest version metadata for Orcheo components."""
    return SystemInfoResponse.model_validate(get_system_info_payload())


@router.get("/system/plugins", response_model=SystemPluginsResponse)
def get_system_plugins() -> SystemPluginsResponse:
    """Return plugin availability for the current backend process."""
    return SystemPluginsResponse.model_validate({"plugins": list_runtime_plugins()})


__all__ = ["public_router", "router"]



"""Concrete in-memory repository implementation."""

from __future__ import annotations
from orcheo.vault.oauth import OAuthCredentialService
from orcheo_backend.app.repository.in_memory.listeners import ListenerRepositoryMixin
from orcheo_backend.app.repository.in_memory.retry import RetryPolicyMixin
from orcheo_backend.app.repository.in_memory.runs import WorkflowRunMixin
from orcheo_backend.app.repository.in_memory.state import InMemoryRepositoryState
from orcheo_backend.app.repository.in_memory.triggers import TriggerDispatchMixin
from orcheo_backend.app.repository.in_memory.versions import WorkflowVersionMixin
from orcheo_backend.app.repository.in_memory.workflows import WorkflowCrudMixin


class InMemoryWorkflowRepository(
    ListenerRepositoryMixin,
    RetryPolicyMixin,
    TriggerDispatchMixin,
    WorkflowRunMixin,
    WorkflowVersionMixin,
    WorkflowCrudMixin,
    InMemoryRepositoryState,
):
    """Simple async-safe in-memory repository for workflows and runs."""

    def __init__(
        self, credential_service: OAuthCredentialService | None = None
    ) -> None:
        """Initialize the repository with an optional credential service."""
        super().__init__(credential_service=credential_service)


__all__ = ["InMemoryWorkflowRepository"]



"""Graph builder module for Orcheo."""

from __future__ import annotations
from collections.abc import Mapping
from typing import Any
from langgraph.graph import StateGraph
from orcheo.graph.ingestion import LANGGRAPH_SCRIPT_FORMAT, load_graph_from_script


class UnsupportedWorkflowGraphFormatError(ValueError):
    """Raised when runtime receives a non-script workflow graph payload."""


def _describe_graph_format(graph_json: Mapping[str, Any]) -> str:
    """Return a human-readable format label for error messages."""
    graph_format = graph_json.get("format")
    if isinstance(graph_format, str) and graph_format.strip():
        return graph_format
    if any(key in graph_json for key in ("nodes", "edges", "edge_nodes")):
        return "legacy-json-graph"
    return "unknown"


def build_graph(graph_json: Mapping[str, Any]) -> StateGraph:
    """Build a LangGraph graph from a configuration payload."""
    if graph_json.get("format") != LANGGRAPH_SCRIPT_FORMAT:
        observed_format = _describe_graph_format(graph_json)
        msg = (
            "Unsupported workflow graph format "
            f"'{observed_format}'. Only '{LANGGRAPH_SCRIPT_FORMAT}' workflow "
            "versions can execute. Re-ingest this workflow from a Python script."
        )
        raise UnsupportedWorkflowGraphFormatError(msg)

    source = graph_json.get("source")
    if not isinstance(source, str) or not source.strip():
        msg = "Script graph configuration requires a non-empty source"
        raise ValueError(msg)
    entrypoint_value = graph_json.get("entrypoint")
    if entrypoint_value is not None and not isinstance(entrypoint_value, str):
        msg = "Entrypoint must be a string when provided"
        raise ValueError(msg)
    return load_graph_from_script(source, entrypoint=entrypoint_value)



"""Workflow download CLI tests for LangGraph source handling."""

from __future__ import annotations
import httpx
import respx
from typer.testing import CliRunner
from orcheo_sdk.cli.main import app


def test_workflow_download_langgraph_script_returns_original_source(
    runner: CliRunner, env: dict[str, str]
) -> None:
    """Test workflow download returns original LangGraph script source."""
    original_source = """from langgraph.graph import StateGraph
from typing import TypedDict

class State(TypedDict):
    messages: list[str]

def my_node(state: State) -> State:
    return {"messages": state["messages"] + ["processed"]}

graph = StateGraph(State)
graph.add_node("my_node", my_node)
graph.set_entry_point("my_node")
graph.set_finish_point("my_node")
"""
    workflow = {"id": "wf-1", "name": "LangGraphWorkflow"}
    versions = [
        {
            "id": "ver-1",
            "version": 1,
            "graph": {
                "format": "langgraph-script",
                "source": original_source,
                "entrypoint": None,
                "index": {"cron": []},
            },
        }
    ]

    with respx.mock(assert_all_called=True) as router:
        router.get("http://api.test/api/workflows/wf-1").mock(
            return_value=httpx.Response(200, json=workflow)
        )
        router.get("http://api.test/api/workflows/wf-1/versions").mock(
            return_value=httpx.Response(200, json=versions)
        )
        result = runner.invoke(app, ["workflow", "download", "wf-1"], env=env)
    assert result.exit_code == 0
    assert "from langgraph.graph import StateGraph" in result.stdout
    assert "def my_node(state: State)" in result.stdout
    # Should NOT contain SDK template code
    assert "from orcheo_sdk import Workflow" not in result.stdout



"""Tests for WebSearchNode Tavily integration."""

from __future__ import annotations
import json
from typing import Any
import httpx
import pytest
import respx
from langchain_core.runnables import RunnableConfig
import orcheo.nodes.conversational_search.retrieval as retrieval_mod
from orcheo.graph.state import State
from orcheo.nodes.conversational_search import WebSearchNode
from orcheo.nodes.conversational_search.models import SearchResult
from orcheo.runtime.credentials import CredentialResolver, credential_resolution
from tests.runtime.credential_test_helpers import create_vault_with_secret


@pytest.mark.asyncio
async def test_web_search_node_requests_tavily_and_formats_results() -> None:
    """WebSearchNode should call Tavily and normalize results."""
    state = State(
        inputs={"query": "latest ai news"},
        results={},
        structured_response=None,
    )
    node = WebSearchNode(
        name="web",
        api_key="tavily-key",
        max_results=3,
        include_raw_content=True,
        search_depth="advanced",
        days=2,
        topic="news",
        include_domains=["example.com"],
        exclude_domains=["ignore.com"],
    )

    captured: dict[str, Any] = {}

    def handler(request: httpx.Request) -> httpx.Response:
        captured["json"] = json.loads(request.content.decode())
        return httpx.Response(
            200,
            json={
                "answer": "summary answer",
                "results": [
                    {
                        "title": "Example Title",
                        "url": "https://example.com",
                        "content": "snippet",
                        "score": 0.9,
                        "raw_content": "full page text",
                    },
                    {"url": "https://example.com/2", "score": "not-a-number"},
                    {"content": "", "raw_content": "raw-only"},
                ],
            },
        )

    with respx.mock(assert_all_called=True) as router:
        router.post("https://api.tavily.com/search").mock(side_effect=handler)
        result = await node(state, RunnableConfig())

    payload = result["results"]["web"]
    request_json = captured["json"]

    assert request_json["api_key"] == "tavily-key"
    assert request_json["query"] == "latest ai news"
    assert request_json["max_results"] == 3
    assert request_json["include_raw_content"] is True
    assert request_json["search_depth"] == "advanced"
    assert request_json["days"] == 2
    assert request_json["topic"] == "news"
    assert request_json["include_domains"] == ["example.com"]
    assert request_json["exclude_domains"] == ["ignore.com"]

    assert payload["answer"] == "summary answer"
    assert payload["source"] == "web"
    raw_results = payload["results"]
    results = [SearchResult.model_validate(entry) for entry in raw_results]
    assert results[0].id == "https://example.com"
    assert results[0].metadata["url"] == "https://example.com"
    assert results[0].metadata["title"] == "Example Title"
    assert results[0].metadata["raw_content"] == "full page text"
    assert results[0].score == pytest.approx(0.9)
    assert results[0].source == "web"
    assert results[1].id == "https://example.com/2"
    assert results[1].score == 0.0
    assert results[1].text == "https://example.com/2"
    assert results[2].id == "web-2"
    assert results[2].metadata == {"raw_content": "raw-only"}
    assert results[2].text == "raw-only"


@pytest.mark.asyncio
async def test_web_search_node_rejects_unknown_provider() -> None:
    node = WebSearchNode(
        name="web", api_key="key", provider="bing", suppress_errors=False
    )
    state = State(
        inputs={"query": "news"},
        results={},
        structured_response=None,
    )

    with pytest.raises(
        ValueError, match="WebSearchNode only supports the 'tavily' provider"
    ):
        await node.run(state, {})


@pytest.mark.asyncio
async def test_web_search_node_requires_query_when_not_suppressed() -> None:
    """WebSearchNode should reject empty queries when suppression is disabled."""
    node = WebSearchNode(name="web", api_key="key", suppress_errors=False)
    state = State(
        inputs={"query": "   "},
        results={},
        structured_response=None,
    )

    with pytest.raises(
        ValueError, match="WebSearchNode requires a non-empty query string"
    ):
        await node.run(state, {})


@pytest.mark.asyncio
async def test_web_search_node_requires_api_key_when_not_suppressed() -> None:
    """WebSearchNode should require api_key when suppression disabled."""
    node = WebSearchNode(name="web", suppress_errors=False)
    state = State(
        inputs={"query": "hello"},
        results={},
        structured_response=None,
    )

    with pytest.raises(
        ValueError, match="WebSearchNode requires api_key to be configured"
    ):
        await node.run(state, {})


@pytest.mark.asyncio
async def test_web_search_node_rejects_whitespace_api_key_when_not_suppressed() -> None:
    """WebSearchNode should treat whitespace api_key as missing."""
    node = WebSearchNode(name="web", api_key="   ", suppress_errors=False)
    state = State(
        inputs={"query": "hello"},
        results={},
        structured_response=None,
    )

    with pytest.raises(
        ValueError, match="WebSearchNode requires api_key to be configured"
    ):
        await node.run(state, {})


@pytest.mark.asyncio
async def test_web_search_node_resolves_api_key_from_credential_vault() -> None:
    """WebSearchNode should resolve api_key placeholders via active resolver."""
    state = State(
        inputs={"query": "latest ai news"},
        results={},
        structured_response=None,
    )
    node = WebSearchNode(
        name="web", api_key="[[telegram_token]]", suppress_errors=False
    )
    resolver = CredentialResolver(create_vault_with_secret(secret="vault-key"))

    captured: dict[str, Any] = {}

    def handler(request: httpx.Request) -> httpx.Response:
        captured["json"] = json.loads(request.content.decode())
        return httpx.Response(200, json={"results": []})

    with respx.mock(assert_all_called=True) as router:
        router.post("https://api.tavily.com/search").mock(side_effect=handler)
        with credential_resolution(resolver):
            await node.run(state, {})

    assert captured["json"]["api_key"] == "vault-key"


def test_web_search_node_raises_for_placeholder_without_active_resolver(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Credential placeholders require an active credential resolver."""
    node = WebSearchNode(name="web", api_key="[[telegram_token]]")
    monkeypatch.setattr(retrieval_mod, "get_active_credential_resolver", lambda: None)

    with pytest.raises(ValueError, match="no active credential resolver is available"):
        node._resolve_api_key()


def test_web_search_node_raises_when_resolved_credential_is_not_string(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Resolved credential values must be strings."""

    class _BadResolver:
        def resolve(self, reference: object) -> int:
            del reference
            return 123

    node = WebSearchNode(name="web", api_key="[[telegram_token]]")
    monkeypatch.setattr(
        retrieval_mod, "get_active_credential_resolver", lambda: _BadResolver()
    )

    with pytest.raises(
        ValueError, match="Resolved WebSearchNode api_key credential must be a string"
    ):
        node._resolve_api_key()


@pytest.mark.asyncio
async def test_web_search_node_handles_http_errors_when_not_suppressed() -> None:
    """WebSearchNode should surface HTTP failures when suppression disabled."""
    node = WebSearchNode(name="web", api_key="key", suppress_errors=False)
    state = State(
        inputs={"query": "hello"},
        results={},
        structured_response=None,
    )

    with respx.mock(assert_all_called=True) as router:
        router.post("https://api.tavily.com/search").mock(
            return_value=httpx.Response(500, text="error")
        )
        with pytest.raises(ValueError, match="Web search request failed"):
            await node.run(state, {})


@pytest.mark.asyncio
async def test_web_search_node_validates_response_shape_when_not_suppressed() -> None:
    """WebSearchNode should validate Tavily payloads when suppression disabled."""
    node = WebSearchNode(name="web", api_key="key", suppress_errors=False)
    state = State(
        inputs={"query": "hello"},
        results={},
        structured_response=None,
    )

    with respx.mock(assert_all_called=True) as router:
        router.post("https://api.tavily.com/search").mock(
            return_value=httpx.Response(200, json={"results": "not-a-list"})
        )
        with pytest.raises(
            ValueError, match="WebSearchNode expected 'results' list in response"
        ):
            await node.run(state, {})

    with respx.mock(assert_all_called=True) as router:
        router.post("https://api.tavily.com/search").mock(
            return_value=httpx.Response(200, json={"results": ["bad-entry"]})
        )
        with pytest.raises(
            ValueError, match="Each web search result must be a mapping"
        ):
            await node.run(state, {})

    with respx.mock(assert_all_called=True) as router:
        router.post("https://api.tavily.com/search").mock(
            return_value=httpx.Response(200, text="not-json")
        )
        with pytest.raises(
            ValueError, match="WebSearchNode received non-JSON response"
        ):
            await node.run(state, {})


@pytest.mark.asyncio
async def test_web_search_node_returns_warning_when_query_missing() -> None:
    """WebSearchNode should return warning payload when query is missing by default."""
    node = WebSearchNode(name="web", api_key="key")
    state = State(
        inputs={"query": "   "},
        results={},
        structured_response=None,
    )

    payload = await node.run(state, {})
    assert payload["results"] == []
    assert "warning" in payload
    assert payload["source"] == "web"


@pytest.mark.asyncio
async def test_web_search_node_suppresses_http_errors_by_default() -> None:
    """WebSearchNode should suppress HTTP errors when suppress_errors is True."""
    node = WebSearchNode(name="web", api_key="key")
    state = State(
        inputs={"query": "hello"},
        results={},
        structured_response=None,
    )

    with respx.mock(assert_all_called=True) as router:
        router.post("https://api.tavily.com/search").mock(
            return_value=httpx.Response(500, text="error")
        )
        payload = await node.run(state, {})

    assert payload["results"] == []
    assert "warning" in payload
    assert payload["source"] == "web"


@pytest.mark.asyncio
async def test_web_search_node_suppresses_unexpected_errors_by_default(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """WebSearchNode should suppress unexpected exceptions when configured."""
    node = WebSearchNode(name="web", api_key="key")
    state = State(
        inputs={"query": "hello"},
        results={},
        structured_response=None,
    )

    async def boom(_: State) -> dict[str, Any]:
        raise RuntimeError("kaboom")

    monkeypatch.setattr(node, "_run_search", boom)
    payload = await node.run(state, {})
    assert payload["results"] == []
    assert payload["warning"] == "web search unavailable: kaboom"
    assert payload["source"] == "web"



"""Filesystem path helpers for the plugin subsystem."""

from __future__ import annotations
import os
from pathlib import Path
from orcheo.plugins.models import PluginStoragePaths


PLUGIN_DIR_ENV = "ORCHEO_PLUGIN_DIR"
CACHE_DIR_ENV = "ORCHEO_CACHE_DIR"


def get_plugin_dir() -> Path:
    """Return the plugin runtime directory."""
    override = os.getenv(PLUGIN_DIR_ENV)
    if override:
        return Path(override).expanduser()
    return Path.home() / ".orcheo" / "plugins"


def get_cache_root() -> Path:
    """Return the Orcheo cache root."""
    override = os.getenv(CACHE_DIR_ENV)
    if override:
        return Path(override).expanduser()
    return Path.home() / ".cache" / "orcheo"


def build_storage_paths() -> PluginStoragePaths:
    """Return the default plugin storage layout."""
    plugin_dir = get_plugin_dir()
    cache_root = get_cache_root()
    cache_dir = cache_root / "plugins"
    return PluginStoragePaths(
        plugin_dir=str(plugin_dir),
        state_file=str(plugin_dir / "plugins.toml"),
        lock_file=str(plugin_dir / "plugin-lock.toml"),
        install_dir=str(plugin_dir / "venv"),
        wheels_dir=str(plugin_dir / "wheels"),
        manifests_dir=str(plugin_dir / "manifests"),
        cache_dir=str(cache_dir),
        downloads_dir=str(cache_dir / "downloads"),
        metadata_dir=str(cache_dir / "metadata"),
    )


__all__ = [
    "CACHE_DIR_ENV",
    "PLUGIN_DIR_ENV",
    "build_storage_paths",
    "get_cache_root",
    "get_plugin_dir",
]



"""Convert LangGraph StateGraph instances into JSON-serialisable summaries."""

from __future__ import annotations
from typing import Any
from langgraph.graph import StateGraph
from langgraph.graph.state import CompiledStateGraph
from pydantic import BaseModel
from orcheo.graph.mermaid import (
    has_workflow_tool_subgraphs,
    normalise_mermaid_sentinels,
    render_summary_mermaid,
)
from orcheo.nodes.registry import registry


def summarise_state_graph(graph: StateGraph) -> dict[str, Any]:
    """Return a JSON-serialisable summary of the ``StateGraph`` structure."""
    nodes = [
        _serialise_node(name, spec.runnable, nested_graph_depth=1)
        for name, spec in graph.nodes.items()
    ]
    edges = [_normalise_edge(edge) for edge in sorted(graph.edges)]
    branches = [
        _serialise_branch(source, branch_name, branch)
        for source, branch_map in graph.branches.items()
        for branch_name, branch in branch_map.items()
    ]

    return {
        "nodes": nodes,
        "edges": edges,
        "conditional_edges": [
            branch
            for branch in branches
            if branch.get("mapping") or branch.get("default")
        ],
    }


def summarise_graph_index(
    graph: StateGraph,
) -> dict[str, Any]:
    """Return compact graph metadata for downstream CLI/UI consumers."""
    summary = summarise_state_graph(graph)
    index: dict[str, Any] = {
        "cron": _extract_cron_index(graph),
        "listeners": _extract_listener_index(graph),
    }
    compact_mermaid = _render_compact_mermaid(graph)
    mermaid = (
        render_summary_mermaid(summary)
        if has_workflow_tool_subgraphs(summary)
        else compact_mermaid
    )
    if mermaid:
        index["mermaid"] = mermaid
    if compact_mermaid and mermaid and compact_mermaid != mermaid:
        index["mermaid_compact"] = compact_mermaid
    return index


def _serialise_node(
    name: str,
    runnable: Any,
    *,
    nested_graph_depth: int = 0,
) -> dict[str, Any]:
    """Return a JSON representation for a LangGraph node."""
    runnable_obj = _unwrap_runnable(runnable)
    metadata = registry.get_metadata_by_callable(runnable_obj)
    node_type = metadata.name if metadata else type(runnable_obj).__name__
    payload = {"name": name, "type": node_type}

    if isinstance(runnable_obj, BaseModel):
        node_config = runnable_obj.model_dump(
            mode="json",
            fallback=lambda value: _serialise_fallback(
                value,
                nested_graph_depth=nested_graph_depth,
            ),
        )
        node_config.pop("name", None)
        payload.update(node_config)

    return payload


def _serialise_fallback(
    value: Any,
    *,
    nested_graph_depth: int = 0,
) -> Any:
    """Return a JSON-safe representation for unsupported Pydantic values."""
    if isinstance(value, BaseModel):
        return value.model_dump(
            mode="json",
            fallback=lambda nested: _serialise_fallback(
                nested,
                nested_graph_depth=nested_graph_depth,
            ),
        )
    if isinstance(value, StateGraph):
        payload: dict[str, Any] = {
            "type": "StateGraph",
            "nodes": sorted(value.nodes.keys()),
        }
        if nested_graph_depth > 0:
            payload["summary"] = summarise_state_graph_with_depth(
                value,
                nested_graph_depth=nested_graph_depth - 1,
            )
        return payload
    if isinstance(value, CompiledStateGraph):
        return {"type": "CompiledStateGraph"}
    if isinstance(value, type):
        return f"{value.__module__}.{value.__name__}"
    if isinstance(value, set):
        return [
            _serialise_fallback(item, nested_graph_depth=nested_graph_depth)
            for item in value
        ]
    return repr(value)


def summarise_state_graph_with_depth(
    graph: StateGraph,
    *,
    nested_graph_depth: int,
) -> dict[str, Any]:
    """Return a graph summary with a bounded nested graph depth."""
    nodes = [
        _serialise_node(name, spec.runnable, nested_graph_depth=nested_graph_depth)
        for name, spec in graph.nodes.items()
    ]
    edges = [_normalise_edge(edge) for edge in sorted(graph.edges)]
    branches = [
        _serialise_branch(source, branch_name, branch)
        for source, branch_map in graph.branches.items()
        for branch_name, branch in branch_map.items()
    ]

    return {
        "nodes": nodes,
        "edges": edges,
        "conditional_edges": [
            branch
            for branch in branches
            if branch.get("mapping") or branch.get("default")
        ],
    }


def _render_compact_mermaid(graph: StateGraph) -> str | None:
    """Return Mermaid text for the top-level graph when rendering succeeds."""
    try:
        return normalise_mermaid_sentinels(graph.compile().get_graph().draw_mermaid())
    except Exception:
        return None


def _extract_cron_index(graph: StateGraph) -> list[dict[str, Any]]:
    """Extract cron trigger fields from ``CronTriggerNode`` graph nodes."""
    cron_nodes: list[dict[str, Any]] = []
    for name, spec in graph.nodes.items():
        node = _serialise_node(name, spec.runnable)
        if node.get("type") != "CronTriggerNode":
            continue

        payload: dict[str, Any] = {}
        for key in (
            "expression",
            "timezone",
            "allow_overlapping",
            "start_at",
            "end_at",
        ):
            if key in node:
                payload[key] = node.get(key)
        cron_nodes.append(payload)
    return cron_nodes


def _extract_listener_index(graph: StateGraph) -> list[dict[str, Any]]:
    """Extract listener node metadata from the graph."""
    from orcheo.nodes.listeners import ListenerNode

    listener_nodes: list[dict[str, Any]] = []
    for name, spec in graph.nodes.items():
        runnable_obj = _unwrap_runnable(spec.runnable)
        if not isinstance(runnable_obj, ListenerNode):
            continue
        node = _serialise_node(name, runnable_obj)

        payload: dict[str, Any] = {
            "node_name": name,
            "platform": node.get("platform"),
            "type": node.get("type"),
        }
        for key, value in node.items():
            if key in {"name", "type", "platform"}:
                continue
            payload[key] = value
        listener_nodes.append(payload)
    return listener_nodes


def _unwrap_runnable(runnable: Any) -> Any:
    """Return the underlying callable stored within LangGraph wrappers."""
    if hasattr(runnable, "afunc") and isinstance(runnable.afunc, BaseModel):
        return runnable.afunc
    if hasattr(runnable, "func") and isinstance(runnable.func, BaseModel):
        return runnable.func
    return runnable


def _serialise_branch(source: str, name: str, branch: Any) -> dict[str, Any]:
    """Return metadata describing a conditional branch."""
    mapping: dict[str, str] | None = None
    ends = getattr(branch, "ends", None)
    if isinstance(ends, dict):
        mapping = {str(key): _normalise_vertex(target) for key, target in ends.items()}

    default: str | None = None
    then_target = getattr(branch, "then", None)
    if isinstance(then_target, str):
        default = _normalise_vertex(then_target)

    payload: dict[str, Any] = {
        "source": source,
        "branch": name,
    }
    if mapping:
        payload["mapping"] = mapping
    if default is not None:
        payload["default"] = default
    if hasattr(branch, "path") and getattr(branch.path, "func", None):
        payload["callable"] = getattr(branch.path.func, "__name__", "<lambda>")

    return payload


def _normalise_edge(edge: tuple[str, str]) -> tuple[str, str]:
    """Convert LangGraph sentinel edge names into public constants."""
    source, target = edge
    return (_normalise_vertex(source), _normalise_vertex(target))


def _normalise_vertex(value: str) -> str:
    """Map LangGraph sentinel vertex names to ``START``/``END``."""
    if value == "__start__":
        return "START"
    if value == "__end__":
        return "END"
    return value


__all__ = ["summarise_graph_index", "summarise_state_graph"]



"""Tool preparation tests for AI nodes."""

from __future__ import annotations
from unittest.mock import AsyncMock, MagicMock, patch
import pytest
from langchain_core.runnables import RunnableConfig
from langchain_core.tools import BaseTool
from orcheo.graph.state import State


@pytest.mark.asyncio
@patch("orcheo.nodes.ai.tool_registry")
@patch("orcheo.nodes.ai.create_agent")
@patch("orcheo.nodes.ai.MultiServerMCPClient")
async def test_prepare_tools(
    mock_mcp_client_class,
    mock_create_agent,
    mock_tool_registry,
    agent,
    mock_agent,
    mock_mcp_client,
):
    """Agent should merge predefined and MCP tools."""

    mock_mcp_client_class.return_value = mock_mcp_client
    mock_mcp_tools = [AsyncMock()]
    mock_mcp_client.get_tools.return_value = mock_mcp_tools
    mock_create_agent.return_value = mock_agent

    mock_tool = MagicMock(spec=BaseTool)
    mock_tool_factory = MagicMock(return_value=mock_tool)
    mock_tool_registry.get_tool.return_value = mock_tool_factory

    agent.predefined_tools = ["tool1"]
    agent.workflow_tools = []
    state: State = {"messages": [{"role": "user", "content": "test"}]}
    config = RunnableConfig()

    await agent.run(state, config)

    mock_mcp_client.get_tools.assert_called_once()
    mock_create_agent.assert_called_once()
    call_kwargs = mock_create_agent.call_args[1]
    assert len(call_kwargs["tools"]) == 2


@pytest.mark.asyncio
@patch("orcheo.nodes.ai.tool_registry")
@patch("orcheo.nodes.ai.create_agent")
@patch("orcheo.nodes.ai.MultiServerMCPClient")
async def test_prepare_tools_with_base_tool_instance(
    mock_mcp_client_class,
    mock_create_agent,
    mock_tool_registry,
    agent,
    mock_agent,
    mock_mcp_client,
):
    """Direct BaseTool instances should be used without wrapping."""

    mock_mcp_client_class.return_value = mock_mcp_client
    mock_mcp_client.get_tools.return_value = []
    mock_create_agent.return_value = mock_agent

    mock_tool = MagicMock(spec=BaseTool)
    mock_tool_registry.get_tool.return_value = mock_tool

    agent.predefined_tools = ["tool1"]
    agent.workflow_tools = []
    state: State = {"messages": [{"role": "user", "content": "test"}]}
    config = RunnableConfig()

    await agent.run(state, config)

    mock_create_agent.assert_called_once()
    call_kwargs = mock_create_agent.call_args[1]
    assert len(call_kwargs["tools"]) == 1
    assert call_kwargs["tools"][0] is mock_tool


@pytest.mark.asyncio
@patch("orcheo.nodes.ai.tool_registry")
@patch("orcheo.nodes.ai.create_agent")
@patch("orcheo.nodes.ai.MultiServerMCPClient")
async def test_prepare_tools_with_none_tool(
    mock_mcp_client_class,
    mock_create_agent,
    mock_tool_registry,
    agent,
    mock_agent,
    mock_mcp_client,
):
    """Missing tools should be skipped gracefully."""

    mock_mcp_client_class.return_value = mock_mcp_client
    mock_mcp_client.get_tools.return_value = []
    mock_create_agent.return_value = mock_agent

    mock_tool_registry.get_tool.return_value = None

    agent.predefined_tools = ["nonexistent_tool"]
    agent.workflow_tools = []
    state: State = {"messages": [{"role": "user", "content": "test"}]}
    config = RunnableConfig()

    await agent.run(state, config)

    mock_create_agent.assert_called_once()
    call_kwargs = mock_create_agent.call_args[1]
    assert len(call_kwargs["tools"]) == 0


@pytest.mark.asyncio
@patch("orcheo.nodes.ai.tool_registry")
@patch("orcheo.nodes.ai.create_agent")
@patch("orcheo.nodes.ai.MultiServerMCPClient")
async def test_prepare_tools_with_non_callable_non_basetool(
    mock_mcp_client_class,
    mock_create_agent,
    mock_tool_registry,
    agent,
    mock_agent,
    mock_mcp_client,
):
    """Invalid registry values should be ignored."""

    mock_mcp_client_class.return_value = mock_mcp_client
    mock_mcp_client.get_tools.return_value = []
    mock_create_agent.return_value = mock_agent

    mock_tool_registry.get_tool.return_value = "not_a_tool"

    agent.predefined_tools = ["invalid_tool"]
    agent.workflow_tools = []
    state: State = {"messages": [{"role": "user", "content": "test"}]}
    config = RunnableConfig()

    await agent.run(state, config)

    mock_create_agent.assert_called_once()
    call_kwargs = mock_create_agent.call_args[1]
    assert len(call_kwargs["tools"]) == 0


@pytest.mark.asyncio
@patch("orcheo.nodes.ai.tool_registry")
@patch("orcheo.nodes.ai.create_agent")
@patch("orcheo.nodes.ai.MultiServerMCPClient")
async def test_prepare_tools_factory_returns_non_basetool(
    mock_mcp_client_class,
    mock_create_agent,
    mock_tool_registry,
    agent,
    mock_agent,
    mock_mcp_client,
):
    """Factories returning non-BaseTool instances should be filtered."""

    mock_mcp_client_class.return_value = mock_mcp_client
    mock_mcp_client.get_tools.return_value = []
    mock_create_agent.return_value = mock_agent

    mock_tool_factory = MagicMock(return_value="not_a_basetool")
    mock_tool_registry.get_tool.return_value = mock_tool_factory

    agent.predefined_tools = ["bad_factory"]
    agent.workflow_tools = []
    state: State = {"messages": [{"role": "user", "content": "test"}]}
    config = RunnableConfig()

    await agent.run(state, config)

    mock_tool_factory.assert_called_once()
    mock_create_agent.assert_called_once()
    call_kwargs = mock_create_agent.call_args[1]
    assert len(call_kwargs["tools"]) == 0


@pytest.mark.asyncio
@patch("orcheo.nodes.ai.tool_registry")
@patch("orcheo.nodes.ai.create_agent")
@patch("orcheo.nodes.ai.MultiServerMCPClient")
async def test_prepare_tools_factory_raises_exception(
    mock_mcp_client_class,
    mock_create_agent,
    mock_tool_registry,
    agent,
    mock_agent,
    mock_mcp_client,
):
    """Factories raising should not break tool assembly."""

    mock_mcp_client_class.return_value = mock_mcp_client
    mock_mcp_client.get_tools.return_value = []
    mock_create_agent.return_value = mock_agent

    mock_tool_factory = MagicMock(side_effect=ValueError("Factory failed"))
    mock_tool_registry.get_tool.return_value = mock_tool_factory

    agent.predefined_tools = ["failing_factory"]
    agent.workflow_tools = []
    state: State = {"messages": [{"role": "user", "content": "test"}]}
    config = RunnableConfig()

    await agent.run(state, config)

    mock_tool_factory.assert_called_once()
    mock_create_agent.assert_called_once()
    call_kwargs = mock_create_agent.call_args[1]
    assert len(call_kwargs["tools"]) == 0



"""ChatKit runtime helpers."""

from __future__ import annotations
import pytest
from orcheo_backend.app import chatkit_runtime


def test_sensitive_logging_enabled_accepts_dev_environment(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Development-like env vars enable sensitive logging."""
    monkeypatch.setenv("ORCHEO_ENV", "DEV")
    monkeypatch.delenv("NODE_ENV", raising=False)
    monkeypatch.delenv("LOG_SENSITIVE_DEBUG", raising=False)

    assert chatkit_runtime.sensitive_logging_enabled() is True


def test_sensitive_logging_enabled_defaults_to_false(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Sensitive logging stays disabled outside known environments."""
    monkeypatch.delenv("ORCHEO_ENV", raising=False)
    monkeypatch.setenv("NODE_ENV", "production")
    monkeypatch.delenv("LOG_SENSITIVE_DEBUG", raising=False)

    assert chatkit_runtime.sensitive_logging_enabled() is False


def test_sensitive_logging_enabled_checks_flag(monkeypatch: pytest.MonkeyPatch) -> None:
    """The LOG_SENSITIVE_DEBUG flag overrides non-dev environments."""
    monkeypatch.delenv("ORCHEO_ENV", raising=False)
    monkeypatch.delenv("NODE_ENV", raising=False)
    monkeypatch.setenv("LOG_SENSITIVE_DEBUG", "1")

    assert chatkit_runtime.sensitive_logging_enabled() is True


def test_get_chatkit_server_initializes_and_caches(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """get_chatkit_server creates a singleton server instance."""
    chatkit_runtime._chatkit_server_ref["server"] = None

    repository = object()
    created = []

    def fake_get_repository() -> object:
        return repository

    def fake_create_chatkit_server(repo_arg, vault_factory):  # type: ignore[no-untyped-def]
        assert repo_arg is repository
        assert vault_factory is chatkit_runtime.get_vault
        server = object()
        created.append(server)
        return server

    monkeypatch.setattr(chatkit_runtime, "get_repository", fake_get_repository)
    monkeypatch.setattr(
        chatkit_runtime, "create_chatkit_server", fake_create_chatkit_server
    )

    try:
        first_server = chatkit_runtime.get_chatkit_server()
        second_server = chatkit_runtime.get_chatkit_server()

        assert first_server is second_server
        assert created == [first_server]
    finally:
        chatkit_runtime._chatkit_server_ref["server"] = None



"""Tests for SlackEventsParserNode."""

from __future__ import annotations
import hashlib
import hmac
import json
import time
from collections.abc import Mapping
from typing import Any
import pytest
from langchain_core.runnables import RunnableConfig
from pydantic import BaseModel
from orcheo.graph.state import State
from orcheo.nodes.slack import SlackEventsParserNode


def _sign_slack(secret: str, timestamp: int, body: str) -> str:
    base = f"v0:{timestamp}:{body}".encode()
    digest = hmac.new(secret.encode("utf-8"), base, hashlib.sha256).hexdigest()
    return f"v0={digest}"


def _build_state(raw_body: str, headers: dict[str, str]) -> State:
    return State(
        messages=[], inputs={"body": {"raw": raw_body}, "headers": headers}, results={}
    )


def _build_script_state(raw_body: str, headers: dict[str, str]) -> dict[str, Any]:
    return {"body": {"raw": raw_body}, "headers": headers}


def _build_mixed_state(raw_body: str, headers: dict[str, str]) -> dict[str, Any]:
    return {
        "inputs": {},
        "body": {"raw": raw_body},
        "headers": headers,
    }


class DummyBaseState(BaseModel):
    inputs: dict[str, Any]


class DummyBodyState(BaseModel):
    body: dict[str, Any]


class RawBodyMapping(Mapping[str, str]):
    def __init__(self, raw: str) -> None:
        self._raw = raw

    def __getitem__(self, key: str) -> str:
        if key == "raw":
            return self._raw
        raise KeyError(key)

    def __iter__(self):
        return iter(("raw",))

    def __len__(self) -> int:
        return 1


def test_timestamp_tolerance_validator_rejects_invalid_values() -> None:
    with pytest.raises(
        ValueError, match="timestamp_tolerance_seconds must be an integer"
    ):
        SlackEventsParserNode(
            name="slack_events_parser",
            signing_secret="slack-secret",
            timestamp_tolerance_seconds="bad",
        )

    with pytest.raises(ValueError, match="timestamp_tolerance_seconds must be >= 0"):
        SlackEventsParserNode(
            name="slack_events_parser",
            signing_secret="slack-secret",
            timestamp_tolerance_seconds=-1,
        )


def test_extract_inputs_handles_base_model_and_mappings() -> None:
    node = SlackEventsParserNode(name="slack_events_parser", signing_secret="secret")

    state = State(
        messages=[],
        inputs={"body": {"raw": "{}"}, "headers": {"x": "y"}},
        results={},
    )
    assert node._extract_inputs(state) == {"body": {"raw": "{}"}, "headers": {"x": "y"}}

    merged_state = {
        "inputs": {"body": {"raw": "{}"}},
        "body": {"raw": "{}"},
        "headers": {"X-Test": "value"},
        "query_params": {"foo": "bar"},
        "source_ip": "127.0.0.1",
    }
    merged = node._extract_inputs(merged_state)
    assert merged["headers"] == {"X-Test": "value"}
    assert merged["query_params"] == {"foo": "bar"}
    assert merged["source_ip"] == "127.0.0.1"

    assert node._extract_inputs(123) == {}


def test_extract_inputs_from_base_model_mapping_inputs() -> None:
    node = SlackEventsParserNode(name="slack_events_parser", signing_secret="secret")
    state = DummyBaseState(
        inputs={"body": {"raw": "{}"}, "headers": {"x-test": "value"}},
    )

    assert node._extract_inputs(state) == {
        "body": {"raw": "{}"},
        "headers": {"x-test": "value"},
    }


def test_extract_inputs_from_base_model_without_inputs() -> None:
    node = SlackEventsParserNode(name="slack_events_parser", signing_secret="secret")
    state = DummyBodyState(body={"raw": "{}"})

    assert node._extract_inputs(state) == {"body": {"raw": "{}"}}


def test_extract_raw_body_handles_supported_formats() -> None:
    node = SlackEventsParserNode(name="slack_events_parser", signing_secret="secret")
    payload = {"type": "event_callback", "event": {"type": "app_mention"}}
    raw = json.dumps(payload, separators=(",", ":"), ensure_ascii=True)

    raw_from_mapping = node._extract_raw_body({"raw": raw})
    assert raw_from_mapping[0] == raw
    assert raw_from_mapping[1] == payload

    raw_from_bytes = node._extract_raw_body(raw.encode("utf-8"))
    assert raw_from_bytes[0] == raw
    assert raw_from_bytes[1] == payload

    raw_from_dict = node._extract_raw_body(payload)
    assert raw_from_dict[0] == json.dumps(
        payload, separators=(",", ":"), ensure_ascii=True
    )
    assert raw_from_dict[1] == payload

    with pytest.raises(
        ValueError, match="Slack event payload must be a dict, string, or bytes"
    ):
        node._extract_raw_body(1234)


def test_extract_raw_body_accepts_custom_mapping() -> None:
    node = SlackEventsParserNode(name="slack_events_parser", signing_secret="secret")
    payload = {"type": "event_callback"}
    raw = json.dumps(payload, separators=(",", ":"), ensure_ascii=True)

    raw_body, parsed = node._extract_raw_body(RawBodyMapping(raw))
    assert raw_body == raw
    assert parsed == payload


def test_parse_json_returns_empty_for_invalid_data() -> None:
    node = SlackEventsParserNode(name="slack_events_parser", signing_secret="secret")
    assert node._parse_json("not json") == {}
    assert node._parse_json(json.dumps(["not", "dict"])) == {}


def test_verify_signature_validates_headers_and_tolerance() -> None:
    secret = "slack-secret"
    timestamp = int(time.time())
    raw_body = json.dumps({"type": "event"}, separators=(",", ":"), ensure_ascii=True)

    node = SlackEventsParserNode(name="slack_events_parser", signing_secret=secret)

    with pytest.raises(ValueError, match="Missing Slack signature headers"):
        node._verify_signature(raw_body, {})

    headers = {
        "x-slack-signature": "v0=bad",
        "x-slack-request-timestamp": "invalid",
    }
    with pytest.raises(ValueError, match="Invalid Slack timestamp header"):
        node._verify_signature(raw_body, headers)

    node.timestamp_tolerance_seconds = "bad"
    valid_headers = {
        "x-slack-signature": _sign_slack(secret, timestamp, raw_body),
        "x-slack-request-timestamp": str(timestamp),
    }
    with pytest.raises(
        ValueError, match="timestamp_tolerance_seconds must resolve to an integer"
    ):
        node._verify_signature(raw_body, valid_headers)

    node.timestamp_tolerance_seconds = 1
    old_timestamp = timestamp - 10
    old_headers = {
        "x-slack-signature": _sign_slack(secret, old_timestamp, raw_body),
        "x-slack-request-timestamp": str(old_timestamp),
    }
    with pytest.raises(
        ValueError, match="Slack request timestamp outside tolerance window"
    ):
        node._verify_signature(raw_body, old_headers)


def test_verify_signature_accepts_valid_headers() -> None:
    secret = "slack-secret"
    timestamp = int(time.time())
    raw_body = json.dumps(
        {"type": "event_callback"}, separators=(",", ":"), ensure_ascii=True
    )

    headers = {
        "x-slack-signature": _sign_slack(secret, timestamp, raw_body),
        "x-slack-request-timestamp": str(timestamp),
    }

    node = SlackEventsParserNode(name="slack_events_parser", signing_secret=secret)
    node._verify_signature(raw_body, headers)


def test_slack_events_parser_allows_templated_tolerance() -> None:
    node = SlackEventsParserNode(
        name="slack_events_parser",
        signing_secret="slack-secret",
        timestamp_tolerance_seconds="{{config.configurable.signature_tolerance_seconds}}",
    )

    assert (
        node.timestamp_tolerance_seconds
        == "{{config.configurable.signature_tolerance_seconds}}"
    )


@pytest.mark.asyncio
async def test_slack_events_parser_accepts_valid_event() -> None:
    secret = "slack-secret"
    timestamp = int(time.time())
    payload = {
        "type": "event_callback",
        "event": {
            "type": "app_mention",
            "channel": "C123",
            "user": "U123",
            "text": "hello",
        },
    }
    raw_body = json.dumps(payload, separators=(",", ":"), ensure_ascii=True)
    headers = {
        "x-slack-signature": _sign_slack(secret, timestamp, raw_body),
        "x-slack-request-timestamp": str(timestamp),
    }

    node = SlackEventsParserNode(
        name="slack_events_parser",
        signing_secret=secret,
        channel_id="C123",
        allowed_event_types=["app_mention"],
    )
    result = await node.run(_build_state(raw_body, headers), RunnableConfig())

    assert result["should_process"] is True
    assert result["event_type"] == "app_mention"
    assert result["channel"] == "C123"
    assert result["user"] == "U123"
    assert result["text"] == "hello"


@pytest.mark.asyncio
async def test_slack_events_parser_accepts_script_state_inputs() -> None:
    secret = "slack-secret"
    timestamp = int(time.time())
    payload = {
        "type": "event_callback",
        "event": {
            "type": "app_mention",
            "channel": "C123",
            "user": "U123",
            "text": "hello",
        },
    }
    raw_body = json.dumps(payload, separators=(",", ":"), ensure_ascii=True)
    headers = {
        "x-slack-signature": _sign_slack(secret, timestamp, raw_body),
        "x-slack-request-timestamp": str(timestamp),
    }

    node = SlackEventsParserNode(
        name="slack_events_parser",
        signing_secret=secret,
        channel_id="C123",
        allowed_event_types=["app_mention"],
    )
    result = await node.run(_build_script_state(raw_body, headers), RunnableConfig())

    assert result["should_process"] is True
    assert result["event_type"] == "app_mention"


@pytest.mark.asyncio
async def test_slack_events_parser_merges_top_level_payload() -> None:
    secret = "slack-secret"
    timestamp = int(time.time())
    payload = {
        "type": "event_callback",
        "event": {
            "type": "app_mention",
            "channel": "C123",
        },
    }
    raw_body = json.dumps(payload, separators=(",", ":"), ensure_ascii=True)
    headers = {
        "x-slack-signature": _sign_slack(secret, timestamp, raw_body),
        "x-slack-request-timestamp": str(timestamp),
    }

    node = SlackEventsParserNode(name="slack_events_parser", signing_secret=secret)
    result = await node.run(_build_mixed_state(raw_body, headers), RunnableConfig())

    assert result["should_process"] is True
    assert result["event_type"] == "app_mention"


@pytest.mark.asyncio
async def test_slack_events_parser_url_verification() -> None:
    secret = "slack-secret"
    timestamp = int(time.time())
    payload = {"type": "url_verification", "challenge": "abc"}
    raw_body = json.dumps(payload, separators=(",", ":"), ensure_ascii=True)
    headers = {
        "x-slack-signature": _sign_slack(secret, timestamp, raw_body),
        "x-slack-request-timestamp": str(timestamp),
    }

    node = SlackEventsParserNode(name="slack_events_parser", signing_secret=secret)
    result = await node.run(_build_state(raw_body, headers), RunnableConfig())

    assert result["is_verification"] is True
    assert result["challenge"] == "abc"
    assert result["should_process"] is False


@pytest.mark.asyncio
async def test_slack_events_parser_channel_filter_blocks() -> None:
    secret = "slack-secret"
    timestamp = int(time.time())
    payload = {
        "type": "event_callback",
        "event": {"type": "app_mention", "channel": "C999"},
    }
    raw_body = json.dumps(payload, separators=(",", ":"), ensure_ascii=True)
    headers = {
        "x-slack-signature": _sign_slack(secret, timestamp, raw_body),
        "x-slack-request-timestamp": str(timestamp),
    }

    node = SlackEventsParserNode(
        name="slack_events_parser",
        signing_secret=secret,
        channel_id="C123",
    )
    result = await node.run(_build_state(raw_body, headers), RunnableConfig())

    assert result["should_process"] is False
    assert result["event_type"] == "app_mention"


@pytest.mark.asyncio
async def test_slack_events_parser_rejects_invalid_signature() -> None:
    secret = "slack-secret"
    timestamp = int(time.time())
    payload = {
        "type": "event_callback",
        "event": {"type": "app_mention", "channel": "C123"},
    }
    raw_body = json.dumps(payload, separators=(",", ":"), ensure_ascii=True)
    headers = {
        "x-slack-signature": _sign_slack("wrong", timestamp, raw_body),
        "x-slack-request-timestamp": str(timestamp),
    }

    node = SlackEventsParserNode(name="slack_events_parser", signing_secret=secret)

    with pytest.raises(ValueError, match="Slack signature verification failed"):
        await node.run(_build_state(raw_body, headers), RunnableConfig())


@pytest.mark.asyncio
async def test_slack_events_parser_rejects_non_dict_headers() -> None:
    node = SlackEventsParserNode(name="slack_events_parser", signing_secret="")
    with pytest.raises(ValueError, match="Slack event headers must be a dictionary"):
        await node.run({"headers": "not a dict"}, RunnableConfig())


@pytest.mark.asyncio
async def test_slack_events_parser_requires_raw_payload_for_signature() -> None:
    node = SlackEventsParserNode(name="slack_events_parser", signing_secret="secret")
    state = {"body": {"raw": ""}, "headers": {}}
    with pytest.raises(
        ValueError, match="Slack signature verification requires raw payload"
    ):
        await node.run(state, RunnableConfig())


@pytest.mark.asyncio
async def test_slack_events_parser_handles_unknown_payload_types() -> None:
    node = SlackEventsParserNode(name="slack_events_parser", signing_secret="")
    payload = {"type": "app_rate_limit", "event": {"type": "app_rate_limit"}}
    raw_body = json.dumps(payload, separators=(",", ":"), ensure_ascii=True)

    result = await node.run({"body": raw_body, "headers": {}}, RunnableConfig())
    assert result["should_process"] is False
    assert result["event_type"] == "app_rate_limit"


@pytest.mark.asyncio
async def test_slack_events_parser_filters_disallowed_event_types() -> None:
    secret = "slack-secret"
    timestamp = int(time.time())
    payload = {
        "type": "event_callback",
        "event": {"type": "reaction_added", "channel": "C123"},
    }
    raw_body = json.dumps(payload, separators=(",", ":"), ensure_ascii=True)
    headers = {
        "x-slack-signature": _sign_slack(secret, timestamp, raw_body),
        "x-slack-request-timestamp": str(timestamp),
    }

    node = SlackEventsParserNode(
        name="slack_events_parser",
        signing_secret=secret,
        allowed_event_types=["app_mention"],
    )
    result = await node.run(_build_state(raw_body, headers), RunnableConfig())
    assert result["should_process"] is False
    assert result["event_type"] == "reaction_added"


@pytest.mark.asyncio
async def test_slack_events_parser_rejects_retry_requests() -> None:
    secret = "slack-secret"
    timestamp = int(time.time())
    payload = {
        "type": "event_callback",
        "event": {
            "type": "app_mention",
            "channel": "C123",
            "user": "U123",
            "text": "hello",
        },
    }
    raw_body = json.dumps(payload, separators=(",", ":"), ensure_ascii=True)
    headers = {
        "x-slack-signature": _sign_slack(secret, timestamp, raw_body),
        "x-slack-request-timestamp": str(timestamp),
        "X-Slack-Retry-Num": "1",
        "X-Slack-Retry-Reason": "http_timeout",
    }

    node = SlackEventsParserNode(
        name="slack_events_parser",
        signing_secret=secret,
        channel_id="C123",
        allowed_event_types=["app_mention"],
    )
    result = await node.run(_build_state(raw_body, headers), RunnableConfig())

    assert result["should_process"] is False
    assert result["event_type"] is None
    assert result["event"] is None


@pytest.mark.asyncio
async def test_slack_events_parser_accepts_first_delivery_without_retry_header() -> (
    None
):
    secret = "slack-secret"
    timestamp = int(time.time())
    payload = {
        "type": "event_callback",
        "event": {
            "type": "app_mention",
            "channel": "C123",
            "user": "U123",
            "text": "hello",
        },
    }
    raw_body = json.dumps(payload, separators=(",", ":"), ensure_ascii=True)
    headers = {
        "x-slack-signature": _sign_slack(secret, timestamp, raw_body),
        "x-slack-request-timestamp": str(timestamp),
    }

    node = SlackEventsParserNode(
        name="slack_events_parser",
        signing_secret=secret,
        channel_id="C123",
        allowed_event_types=["app_mention"],
    )
    result = await node.run(_build_state(raw_body, headers), RunnableConfig())

    assert result["should_process"] is True
    assert result["event_type"] == "app_mention"



"""Tests for compiler helpers that build listener subscriptions."""

from __future__ import annotations
from uuid import uuid4
from orcheo.listeners.compiler import compile_listener_subscriptions
from orcheo.listeners.models import ListenerPlatform


def test_compile_listener_subscriptions_filters_invalid_entries() -> None:
    workflow_id = uuid4()
    version_id = uuid4()
    graph = {
        "index": {
            "listeners": [
                "not a dict",
                {"node_name": "", "platform": "discord"},
                {"node_name": "invalid", "platform": "unknown"},
                {
                    "node_name": "explicit",
                    "platform": "discord",
                    "bot_identity_key": "  explicit-key  ",
                    "extra": "value",
                    "type": "ignored",
                    "name": "main",
                },
            ]
        }
    }

    subscriptions = compile_listener_subscriptions(workflow_id, version_id, graph)

    assert len(subscriptions) == 1
    subscription = subscriptions[0]
    assert subscription.platform == ListenerPlatform.DISCORD
    assert subscription.bot_identity_key == "explicit-key"
    assert subscription.node_name == "explicit"
    assert subscription.config == {
        "extra": "value",
        "bot_identity_key": "  explicit-key  ",
    }


def test_compile_listener_subscriptions_derives_identity_keys() -> None:
    workflow_id = uuid4()
    version_id = uuid4()
    graph = {
        "index": {
            "listeners": [
                {
                    "node_name": "token",
                    "platform": " telegram ",
                    "token": "  tok  ",
                    "extra": "keep",
                },
                {
                    "name": "app-node",
                    "platform": "qq",
                    "app_id": "app-1",
                },
                {
                    "node_name": "credential",
                    "platform": "discord",
                    "credential_ref": " cred-1 ",
                    "payload": "info",
                },
                {
                    "node_name": "fallback",
                    "platform": "discord",
                },
            ]
        }
    }

    subscriptions = compile_listener_subscriptions(workflow_id, version_id, graph)
    nodes = {subscription.node_name: subscription for subscription in subscriptions}

    assert nodes["token"].bot_identity_key == "telegram:tok"
    assert nodes["token"].config["token"] == "  tok  "
    assert nodes["token"].config["extra"] == "keep"
    assert nodes["app-node"].bot_identity_key == "qq:app-1"
    assert nodes["credential"].bot_identity_key == "discord:cred-1"
    assert nodes["credential"].config == {
        "credential_ref": " cred-1 ",
        "payload": "info",
    }
    assert nodes["fallback"].bot_identity_key == "discord:fallback"


def test_compile_listener_subscriptions_returns_empty_for_non_listeners() -> None:
    workflow_id = uuid4()
    version_id = uuid4()
    graph = {"index": {"listeners": "invalid"}}

    assert compile_listener_subscriptions(workflow_id, version_id, graph) == []


def test_compile_listener_subscriptions_skips_empty_platform_value() -> None:
    """Entries with empty/blank platform are skipped (line 29)."""
    workflow_id = uuid4()
    version_id = uuid4()
    graph = {
        "index": {
            "listeners": [
                {"node_name": "bot", "platform": ""},
                {"node_name": "bot2"},  # no platform key at all
                {"node_name": "bot3", "platform": "   "},  # whitespace only
            ]
        }
    }

    result = compile_listener_subscriptions(workflow_id, version_id, graph)
    assert result == []



"""Credential encryption primitives and helpers."""

from __future__ import annotations
import hashlib
import os
from base64 import b64decode, b64encode, urlsafe_b64encode
from typing import Protocol
from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from orcheo.models.base import OrcheoBaseModel


__all__ = [
    "AesGcmCredentialCipher",
    "CredentialCipher",
    "EncryptionEnvelope",
    "FernetCredentialCipher",
]


class CredentialCipher(Protocol):
    """Protocol describing encryption strategies for credential secrets."""

    algorithm: str
    key_id: str

    def encrypt(self, plaintext: str) -> EncryptionEnvelope:
        """Return an envelope containing ciphertext for the plaintext secret."""
        ...

    def decrypt(self, envelope: EncryptionEnvelope) -> str:
        """Decrypt the provided envelope and return the plaintext secret."""
        ...


class EncryptionEnvelope(OrcheoBaseModel):
    """Encrypted payload metadata produced by a :class:`CredentialCipher`."""

    algorithm: str
    key_id: str
    ciphertext: str

    def decrypt(self, cipher: CredentialCipher) -> str:
        """Use the provided cipher to decrypt the envelope."""
        if cipher.algorithm != self.algorithm:
            msg = "Cipher algorithm mismatch during decryption."
            raise ValueError(msg)
        if cipher.key_id != self.key_id:
            msg = "Cipher key identifier mismatch during decryption."
            raise ValueError(msg)
        return cipher.decrypt(self)


class FernetCredentialCipher:
    """Credential cipher that leverages Fernet symmetric encryption."""

    algorithm: str = "fernet.v1"

    def __init__(self, *, key: str, key_id: str = "primary") -> None:
        """Derive a Fernet key from the provided secret string."""
        digest = hashlib.sha256(key.encode("utf-8")).digest()
        derived_key = urlsafe_b64encode(digest)
        self._fernet = Fernet(derived_key)
        self.key_id = key_id

    def encrypt(self, plaintext: str) -> EncryptionEnvelope:
        """Encrypt plaintext credentials and return an envelope."""
        token = self._fernet.encrypt(plaintext.encode("utf-8"))
        return EncryptionEnvelope(
            algorithm=self.algorithm,
            key_id=self.key_id,
            ciphertext=token.decode("utf-8"),
        )

    def decrypt(self, envelope: EncryptionEnvelope) -> str:
        """Decrypt an envelope previously produced by :meth:`encrypt`."""
        try:
            plaintext = self._fernet.decrypt(envelope.ciphertext.encode("utf-8"))
        except InvalidToken as exc:  # pragma: no cover - defensive
            msg = "Unable to decrypt credential payload with provided key."
            raise ValueError(msg) from exc
        return plaintext.decode("utf-8")


class AesGcmCredentialCipher:
    """Credential cipher backed by AES-256 GCM."""

    algorithm: str = "aes256-gcm.v1"

    def __init__(self, *, key: str, key_id: str = "primary") -> None:
        """Derive a 256-bit AES key from the provided secret."""
        digest = hashlib.sha256(key.encode("utf-8")).digest()
        self._aesgcm = AESGCM(digest)
        self.key_id = key_id

    def encrypt(self, plaintext: str) -> EncryptionEnvelope:
        """Encrypt plaintext and return an envelope with nonce+ciphertext."""
        nonce = os.urandom(12)
        ciphertext = self._aesgcm.encrypt(nonce, plaintext.encode("utf-8"), None)
        payload = b64encode(nonce + ciphertext).decode("utf-8")
        return EncryptionEnvelope(
            algorithm=self.algorithm,
            key_id=self.key_id,
            ciphertext=payload,
        )

    def decrypt(self, envelope: EncryptionEnvelope) -> str:
        """Decrypt an :class:`EncryptionEnvelope` produced by this cipher."""
        try:
            decoded = b64decode(envelope.ciphertext.encode("utf-8"))
        except Exception as exc:  # pragma: no cover - defensive
            msg = "Encrypted payload is not valid base64 data."
            raise ValueError(msg) from exc

        if len(decoded) < 12:
            msg = "Encrypted payload is too short to contain a nonce."
            raise ValueError(msg)

        nonce = decoded[:12]
        ciphertext = decoded[12:]
        try:
            plaintext = self._aesgcm.decrypt(nonce, ciphertext, None)
        except Exception as exc:  # pragma: no cover - defensive
            msg = "Unable to decrypt credential payload with provided key."
            raise ValueError(msg) from exc
        return plaintext.decode("utf-8")



"""Public plugin subsystem exports."""

from orcheo.plugins.api import PluginAPI
from orcheo.plugins.loader import (
    PluginLoadReport,
    PluginLoadResult,
    ensure_plugins_loaded,
    invalidate_plugin_loader,
    load_enabled_plugins,
    reset_plugin_loader_for_tests,
)
from orcheo.plugins.manager import PLUGIN_ENTRYPOINT_GROUP, PluginError, PluginManager
from orcheo.plugins.models import (
    PLUGIN_API_VERSION,
    DesiredPluginRecord,
    DoctorCheck,
    DoctorReport,
    LockedPluginRecord,
    PluginImpactSummary,
    PluginManifest,
    PluginStoragePaths,
)
from orcheo.plugins.paths import build_storage_paths, get_cache_root, get_plugin_dir
from orcheo.plugins.state import (
    load_desired_state,
    load_lock_state,
    save_desired_state,
    save_lock_state,
)


__all__ = [
    "PLUGIN_API_VERSION",
    "PLUGIN_ENTRYPOINT_GROUP",
    "PluginAPI",
    "DesiredPluginRecord",
    "DoctorCheck",
    "DoctorReport",
    "LockedPluginRecord",
    "PluginError",
    "PluginImpactSummary",
    "PluginLoadReport",
    "PluginLoadResult",
    "PluginManager",
    "PluginManifest",
    "PluginStoragePaths",
    "build_storage_paths",
    "ensure_plugins_loaded",
    "get_cache_root",
    "get_plugin_dir",
    "invalidate_plugin_loader",
    "load_enabled_plugins",
    "load_desired_state",
    "load_lock_state",
    "reset_plugin_loader_for_tests",
    "save_desired_state",
    "save_lock_state",
]



"""In-memory browser context store with TTL eviction and focus-priority."""

from __future__ import annotations
import threading
from dataclasses import dataclass
from datetime import UTC, datetime, timedelta


@dataclass(slots=True)
class BrowserContextEntry:
    """A single browser tab's context."""

    session_id: str
    page: str
    workflow_id: str | None
    workflow_name: str | None
    focused: bool
    last_seen: datetime
    last_focused_at: datetime | None = None


class BrowserContextStore:
    """In-memory context store keyed by session_id with TTL eviction.

    Thread-safe. Entries are evicted after ``ttl_seconds`` without an update.
    Focus-priority resolution returns the session with the most recent
    ``last_focused_at``, falling back to the most recently seen session.
    """

    def __init__(self, ttl_seconds: int = 300) -> None:
        """Create a store with the given TTL in seconds."""
        self._ttl = timedelta(seconds=ttl_seconds)
        self._sessions: dict[str, BrowserContextEntry] = {}
        self._lock = threading.Lock()
        self._cleanup_timer: threading.Timer | None = None
        self._start_periodic_cleanup()

    def upsert(
        self,
        *,
        session_id: str,
        page: str,
        workflow_id: str | None,
        workflow_name: str | None,
        focused: bool,
        timestamp: datetime | None = None,
    ) -> None:
        """Insert or update a session entry. Resets the TTL."""
        now = timestamp or datetime.now(UTC)
        with self._lock:
            existing = self._sessions.get(session_id)
            last_focused_at = existing.last_focused_at if existing else None
            if focused:
                last_focused_at = now
            self._sessions[session_id] = BrowserContextEntry(
                session_id=session_id,
                page=page,
                workflow_id=workflow_id,
                workflow_name=workflow_name,
                focused=focused,
                last_seen=now,
                last_focused_at=last_focused_at,
            )

    def _start_periodic_cleanup(self) -> None:
        """Schedule a background cleanup every TTL interval."""
        interval = self._ttl.total_seconds()
        self._cleanup_timer = threading.Timer(interval, self._periodic_cleanup)
        self._cleanup_timer.daemon = True
        self._cleanup_timer.start()

    def _periodic_cleanup(self) -> None:
        """Run eviction and reschedule the next cleanup."""
        now = datetime.now(UTC)
        with self._lock:
            self._evict(now)
        self._start_periodic_cleanup()

    def _evict(self, now: datetime) -> None:
        """Remove entries older than TTL. Must be called under lock."""
        cutoff = now - self._ttl
        expired = [
            sid for sid, entry in self._sessions.items() if entry.last_seen < cutoff
        ]
        for sid in expired:
            del self._sessions[sid]

    def get_active(self) -> dict[str, object]:
        """Return the active context using focus-priority resolution.

        Returns the session with the most recent ``last_focused_at``.
        If no session has focus history, returns the most recently seen.
        If no sessions exist, returns a null context.
        """
        now = datetime.now(UTC)
        with self._lock:
            self._evict(now)
            if not self._sessions:
                return {
                    "session_id": None,
                    "page": None,
                    "workflow_id": None,
                    "workflow_name": None,
                    "focused": False,
                    "last_focused_at": None,
                    "staleness_seconds": 0,
                    "total_sessions": 0,
                }

            total = len(self._sessions)

            # Focus-priority: pick session with most recent last_focused_at
            focused_entries = [
                e for e in self._sessions.values() if e.last_focused_at is not None
            ]
            if focused_entries:
                best = max(
                    focused_entries,
                    key=lambda e: e.last_focused_at or datetime.min.replace(tzinfo=UTC),
                )
            else:
                best = max(self._sessions.values(), key=lambda e: e.last_seen)

            staleness = (now - best.last_seen).total_seconds()
            return {
                "session_id": best.session_id,
                "page": best.page,
                "workflow_id": best.workflow_id,
                "workflow_name": best.workflow_name,
                "focused": best.focused,
                "last_focused_at": (
                    best.last_focused_at.isoformat() if best.last_focused_at else None
                ),
                "staleness_seconds": int(staleness),
                "total_sessions": total,
            }

    def get_all_sessions(self) -> list[dict[str, object]]:
        """Return all active (non-expired) sessions."""
        now = datetime.now(UTC)
        with self._lock:
            self._evict(now)
            results: list[dict[str, object]] = []
            for entry in sorted(
                self._sessions.values(), key=lambda e: e.last_seen, reverse=True
            ):
                staleness = (now - entry.last_seen).total_seconds()
                results.append(
                    {
                        "session_id": entry.session_id,
                        "page": entry.page,
                        "workflow_id": entry.workflow_id,
                        "workflow_name": entry.workflow_name,
                        "focused": entry.focused,
                        "last_seen": entry.last_seen.isoformat(),
                        "last_focused_at": (
                            entry.last_focused_at.isoformat()
                            if entry.last_focused_at
                            else None
                        ),
                        "staleness_seconds": int(staleness),
                    }
                )
            return results



"""Tests for workflow version retrieval and diff endpoints."""

from __future__ import annotations
from datetime import UTC, datetime
from uuid import uuid4
import pytest
from fastapi import HTTPException
from orcheo.models.workflow import WorkflowVersion
from orcheo_backend.app import (
    diff_workflow_versions,
    get_workflow_version,
    list_workflow_versions,
)
from orcheo_backend.app.repository import (
    WorkflowNotFoundError,
    WorkflowVersionNotFoundError,
)
from orcheo_backend.app.routers import workflows as workflow_router


@pytest.mark.asyncio()
async def test_list_workflow_versions_success() -> None:
    """List workflow versions endpoint returns versions."""

    workflow_id = uuid4()
    version1_id = uuid4()
    version2_id = uuid4()

    class Repository:
        async def resolve_workflow_ref(self, workflow_ref, *, include_archived=True):
            del workflow_ref, include_archived
            return workflow_id

        async def list_versions(self, wf_id):
            return [
                WorkflowVersion(
                    id=version1_id,
                    workflow_id=wf_id,
                    version=1,
                    graph={},
                    created_by="admin",
                    created_at=datetime.now(tz=UTC),
                    updated_at=datetime.now(tz=UTC),
                ),
                WorkflowVersion(
                    id=version2_id,
                    workflow_id=wf_id,
                    version=2,
                    graph={},
                    created_by="admin",
                    created_at=datetime.now(tz=UTC),
                    updated_at=datetime.now(tz=UTC),
                ),
            ]

    result = await list_workflow_versions(str(workflow_id), Repository())

    assert len(result) == 2
    assert result[0].id == version1_id
    assert result[1].id == version2_id
    assert isinstance(result[0].mermaid, str)
    assert isinstance(result[1].mermaid, str)


@pytest.mark.asyncio()
async def test_list_workflow_versions_not_found() -> None:
    """List workflow versions raises 404 for missing workflow."""

    workflow_id = uuid4()

    class Repository:
        async def resolve_workflow_ref(self, workflow_ref, *, include_archived=True):
            del workflow_ref, include_archived
            return workflow_id

        async def list_versions(self, wf_id):
            raise WorkflowNotFoundError("not found")

    with pytest.raises(HTTPException) as exc_info:
        await list_workflow_versions(str(workflow_id), Repository())

    assert exc_info.value.status_code == 404


@pytest.mark.asyncio()
async def test_get_workflow_version_success() -> None:
    """Get workflow version endpoint returns specific version."""

    workflow_id = uuid4()
    version_id = uuid4()

    class Repository:
        async def resolve_workflow_ref(self, workflow_ref, *, include_archived=True):
            del workflow_ref, include_archived
            return workflow_id

        async def get_version_by_number(self, wf_id, version_number):
            return WorkflowVersion(
                id=version_id,
                workflow_id=wf_id,
                version=version_number,
                graph={},
                created_by="admin",
                created_at=datetime.now(tz=UTC),
                updated_at=datetime.now(tz=UTC),
            )

    result = await get_workflow_version(str(workflow_id), 1, Repository())

    assert result.id == version_id
    assert result.version == 1
    assert isinstance(result.mermaid, str)


@pytest.mark.asyncio()
async def test_get_workflow_version_workflow_not_found() -> None:
    """Get workflow version raises 404 for missing workflow."""

    workflow_id = uuid4()

    class Repository:
        async def resolve_workflow_ref(self, workflow_ref, *, include_archived=True):
            del workflow_ref, include_archived
            return workflow_id

        async def get_version_by_number(self, wf_id, version_number):
            raise WorkflowNotFoundError("not found")

    with pytest.raises(HTTPException) as exc_info:
        await get_workflow_version(str(workflow_id), 1, Repository())

    assert exc_info.value.status_code == 404


@pytest.mark.asyncio()
async def test_get_workflow_version_version_not_found() -> None:
    """Get workflow version raises 404 for missing version."""

    workflow_id = uuid4()

    class Repository:
        async def resolve_workflow_ref(self, workflow_ref, *, include_archived=True):
            del workflow_ref, include_archived
            return workflow_id

        async def get_version_by_number(self, wf_id, version_number):
            raise WorkflowVersionNotFoundError("not found")

    with pytest.raises(HTTPException) as exc_info:
        await get_workflow_version(str(workflow_id), 1, Repository())

    assert exc_info.value.status_code == 404


@pytest.mark.asyncio()
async def test_diff_workflow_versions_success() -> None:
    """Diff workflow versions endpoint returns diff."""

    workflow_id = uuid4()

    class Diff:
        base_version = 1
        target_version = 2
        diff = ["+ node1", "- node2"]

    class Repository:
        async def resolve_workflow_ref(self, workflow_ref, *, include_archived=True):
            del workflow_ref, include_archived
            return workflow_id

        async def diff_versions(self, wf_id, base, target):
            return Diff()

    result = await diff_workflow_versions(str(workflow_id), 1, 2, Repository())

    assert result.base_version == 1
    assert result.target_version == 2
    assert result.diff == ["+ node1", "- node2"]


@pytest.mark.asyncio()
async def test_list_workflow_versions_handles_mermaid_render_failure(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """List workflow versions falls back to null Mermaid if rendering fails."""

    workflow_id = uuid4()
    version_id = uuid4()

    class Repository:
        async def resolve_workflow_ref(self, workflow_ref, *, include_archived=True):
            del workflow_ref, include_archived
            return workflow_id

        async def list_versions(self, wf_id):
            return [
                WorkflowVersion(
                    id=version_id,
                    workflow_id=wf_id,
                    version=1,
                    graph={},
                    created_by="admin",
                    created_at=datetime.now(tz=UTC),
                    updated_at=datetime.now(tz=UTC),
                ),
            ]

    def _raise_mermaid_failure(_graph: object) -> str:
        raise ValueError("boom")

    monkeypatch.setattr(workflow_router, "_mermaid_from_graph", _raise_mermaid_failure)

    result = await list_workflow_versions(str(workflow_id), Repository())

    assert len(result) == 1
    assert result[0].mermaid is None


@pytest.mark.asyncio()
async def test_diff_workflow_versions_workflow_not_found() -> None:
    """Diff workflow versions raises 404 for missing workflow."""

    workflow_id = uuid4()

    class Repository:
        async def resolve_workflow_ref(self, workflow_ref, *, include_archived=True):
            del workflow_ref, include_archived
            return workflow_id

        async def diff_versions(self, wf_id, base, target):
            raise WorkflowNotFoundError("not found")

    with pytest.raises(HTTPException) as exc_info:
        await diff_workflow_versions(str(workflow_id), 1, 2, Repository())

    assert exc_info.value.status_code == 404


@pytest.mark.asyncio()
async def test_diff_workflow_versions_version_not_found() -> None:
    """Diff workflow versions raises 404 for missing version."""

    workflow_id = uuid4()

    class Repository:
        async def resolve_workflow_ref(self, workflow_ref, *, include_archived=True):
            del workflow_ref, include_archived
            return workflow_id

        async def diff_versions(self, wf_id, base, target):
            raise WorkflowVersionNotFoundError("not found")

    with pytest.raises(HTTPException) as exc_info:
        await diff_workflow_versions(str(workflow_id), 1, 2, Repository())

    assert exc_info.value.status_code == 404


def test_attach_mermaid_non_mapping_graph_uses_renderer(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """_attach_mermaid should fall back to renderer when graph is not a mapping."""

    version = WorkflowVersion(
        id=uuid4(),
        workflow_id=uuid4(),
        version=1,
        graph={},
        created_by="admin",
        created_at=datetime.now(tz=UTC),
        updated_at=datetime.now(tz=UTC),
    )
    object.__setattr__(version, "graph", "serialized-graph")
    monkeypatch.setattr(
        workflow_router,
        "_mermaid_from_graph",
        lambda _graph: "flowchart TD; A-->B",
    )

    result = workflow_router._attach_mermaid(version)

    assert result.mermaid == "flowchart TD; A-->B"



"""SQLite-backed credential vault implementation."""

from __future__ import annotations
from pathlib import Path
from orcheo.models import CredentialCipher
from orcheo.vault.base import BaseCredentialVault
from orcheo.vault.sqlite_storage import SQLiteCredentialVaultMixin


class FileCredentialVault(SQLiteCredentialVaultMixin, BaseCredentialVault):
    """File-backed credential vault stored in a SQLite database."""

    def __init__(
        self, path: str | Path, *, cipher: CredentialCipher | None = None
    ) -> None:
        """Initialize the SQLite-backed vault."""
        BaseCredentialVault.__init__(self, cipher=cipher)
        SQLiteCredentialVaultMixin.__init__(self, path=path)


__all__ = ["FileCredentialVault"]



"""Tests for ingestion helper utilities that serialise graph metadata."""

from __future__ import annotations
import asyncio
from types import SimpleNamespace
import pytest
from langgraph.graph import StateGraph
from orcheo.graph.ingestion import _resolve_graph, _serialise_branch, _unwrap_runnable
from orcheo.nodes.rss import RSSNode


def test_unwrap_runnable_prefers_wrapped_func() -> None:
    node = RSSNode(name="rss", sources=["https://example.com/feed"])
    wrapper = SimpleNamespace(func=node)

    assert _unwrap_runnable(wrapper) is node


def test_serialise_branch_with_mapping_and_default() -> None:
    branch = SimpleNamespace(
        ends={"success": "__start__", "failure": "__end__"},
        then="__end__",
        path=SimpleNamespace(func=lambda: None),
    )

    payload = _serialise_branch("node", "result", branch)

    assert payload["mapping"] == {"success": "START", "failure": "END"}
    assert payload["default"] == "END"
    assert payload["callable"] == "<lambda>"


def test_serialise_branch_without_optional_fields() -> None:
    branch = SimpleNamespace(ends=None, then=None)

    payload = _serialise_branch("node", "result", branch)

    assert payload == {"source": "node", "branch": "result"}


def test_resolve_graph_with_unknown_object_returns_none() -> None:
    assert _resolve_graph(object()) is None


def test_resolve_graph_handles_asyncio_run_runtime_error(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    async def build_graph() -> StateGraph:
        return StateGraph(dict)

    def fake_asyncio_run(*_args: object, **_kwargs: object) -> None:
        raise RuntimeError("boom")

    monkeypatch.setattr(asyncio, "run", fake_asyncio_run)

    graph = _resolve_graph(build_graph())

    assert isinstance(graph, StateGraph)


@pytest.mark.asyncio
async def test_resolve_graph_handles_running_event_loop() -> None:
    async def build_graph() -> StateGraph:
        return StateGraph(dict)

    graph = _resolve_graph(build_graph())

    assert isinstance(graph, StateGraph)



"""Persistence backends for Agentensor training checkpoints."""

from __future__ import annotations
import asyncio
import importlib
import json
from collections.abc import AsyncIterator, Mapping
from contextlib import asynccontextmanager
from datetime import datetime
from pathlib import Path
from typing import Any
import aiosqlite
from orcheo.agentensor.checkpoints import (
    AgentensorCheckpoint,
    AgentensorCheckpointNotFoundError,
    AgentensorCheckpointStore,
)
from orcheo_backend.app.history.sqlite_utils import (
    connect_sqlite,
    ensure_sqlite_schema,
)


# Optional psycopg dependencies
_AsyncConnectionPool: Any | None
_DictRowFactory: Any | None

try:  # pragma: no cover - optional dependency
    _AsyncConnectionPool = importlib.import_module("psycopg_pool").AsyncConnectionPool
    _DictRowFactory = importlib.import_module("psycopg.rows").dict_row
except Exception:  # pragma: no cover - fallback when dependency missing
    _AsyncConnectionPool = None
    _DictRowFactory = None


POSTGRES_CHECKPOINT_MIGRATION = """
CREATE TABLE IF NOT EXISTS agentensor_checkpoints (
    id TEXT PRIMARY KEY,
    workflow_id TEXT NOT NULL,
    config_version INTEGER NOT NULL,
    runnable_config JSONB NOT NULL,
    metrics JSONB NOT NULL,
    metadata JSONB NOT NULL DEFAULT '{}'::jsonb,
    artifact_url TEXT NULL,
    is_best BOOLEAN NOT NULL DEFAULT FALSE,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW()
);
CREATE INDEX IF NOT EXISTS idx_agentensor_checkpoints_workflow
    ON agentensor_checkpoints (workflow_id, config_version);
CREATE INDEX IF NOT EXISTS idx_agentensor_checkpoints_best
    ON agentensor_checkpoints (workflow_id, is_best);
CREATE INDEX IF NOT EXISTS idx_agentensor_checkpoints_metrics
    ON agentensor_checkpoints USING GIN (metrics);
CREATE INDEX IF NOT EXISTS idx_agentensor_checkpoints_metadata
    ON agentensor_checkpoints USING GIN (metadata);
"""


class InMemoryAgentensorCheckpointStore(AgentensorCheckpointStore):
    """Async-safe in-memory checkpoint store."""

    def __init__(self) -> None:
        """Initialize the in-memory store."""
        self._lock = asyncio.Lock()
        self._checkpoints: dict[str, AgentensorCheckpoint] = {}
        self._by_workflow: dict[str, list[str]] = {}

    async def record_checkpoint(
        self,
        *,
        workflow_id: str,
        runnable_config: Mapping[str, object],
        metrics: Mapping[str, object],
        metadata: Mapping[str, object] | None = None,
        artifact_url: str | None = None,
        is_best: bool = False,
        config_version: int | None = None,
    ) -> AgentensorCheckpoint:
        """Persist a checkpoint and return the stored record."""
        async with self._lock:
            version = config_version or self._next_version(workflow_id)
            checkpoint = AgentensorCheckpoint(
                workflow_id=workflow_id,
                config_version=version,
                runnable_config=dict(runnable_config),
                metrics=dict(metrics),
                metadata=dict(metadata or {}),
                artifact_url=artifact_url,
                is_best=is_best,
            )
            self._checkpoints[checkpoint.id] = checkpoint
            self._by_workflow.setdefault(workflow_id, []).append(checkpoint.id)
            if is_best:
                self._clear_other_best(workflow_id, checkpoint.id)
            return checkpoint

    async def list_checkpoints(
        self,
        workflow_id: str,
        *,
        limit: int | None = None,
    ) -> list[AgentensorCheckpoint]:
        """Return checkpoints for the workflow ordered newest-first."""
        async with self._lock:
            identifiers = list(reversed(self._by_workflow.get(workflow_id, [])))
            if limit is not None:
                identifiers = identifiers[:limit]
            return [self._checkpoints[identifier] for identifier in identifiers]

    async def get_checkpoint(self, checkpoint_id: str) -> AgentensorCheckpoint:
        """Return the checkpoint by identifier or raise when missing."""
        async with self._lock:
            checkpoint = self._checkpoints.get(checkpoint_id)
            if checkpoint is None:
                msg = f"Checkpoint {checkpoint_id!r} not found."
                raise AgentensorCheckpointNotFoundError(msg)
            return checkpoint

    async def latest_checkpoint(
        self,
        workflow_id: str,
    ) -> AgentensorCheckpoint | None:
        """Return the most recent checkpoint for the workflow if present."""
        async with self._lock:
            identifiers = self._by_workflow.get(workflow_id, [])
            if not identifiers:
                return None
            return self._checkpoints[identifiers[-1]]

    def _next_version(self, workflow_id: str) -> int:
        identifiers = self._by_workflow.get(workflow_id, [])
        if not identifiers:
            return 1
        latest = self._checkpoints[identifiers[-1]]
        return latest.config_version + 1

    def _clear_other_best(self, workflow_id: str, checkpoint_id: str) -> None:
        for identifier in self._by_workflow.get(workflow_id, []):
            if identifier == checkpoint_id:
                continue
            checkpoint = self._checkpoints.get(identifier)
            if checkpoint is None:
                continue
            checkpoint.is_best = False


class SqliteAgentensorCheckpointStore(AgentensorCheckpointStore):
    """SQLite-backed checkpoint store shared across backend workers."""

    def __init__(self, database_path: str | Path) -> None:
        """Initialize the SQLite store with the given database path."""
        self._database_path = Path(database_path).expanduser()
        self._lock = asyncio.Lock()
        self._init_lock = asyncio.Lock()
        self._initialized = False

    async def record_checkpoint(
        self,
        *,
        workflow_id: str,
        runnable_config: Mapping[str, object],
        metrics: Mapping[str, object],
        metadata: Mapping[str, object] | None = None,
        artifact_url: str | None = None,
        is_best: bool = False,
        config_version: int | None = None,
    ) -> AgentensorCheckpoint:
        """Persist a checkpoint row and return the stored domain object."""
        await self._ensure_initialized()
        async with self._lock:
            async with connect_sqlite(self._database_path) as conn:
                await conn.execute("BEGIN")
                try:
                    version = await self._resolve_version(
                        conn, workflow_id, config_version
                    )
                    checkpoint = AgentensorCheckpoint(
                        workflow_id=workflow_id,
                        config_version=version,
                        runnable_config=dict(runnable_config),
                        metrics=dict(metrics),
                        metadata=dict(metadata or {}),
                        artifact_url=artifact_url,
                        is_best=is_best,
                    )
                    await conn.execute(
                        """
                        INSERT INTO agentensor_checkpoints (
                            id,
                            workflow_id,
                            config_version,
                            runnable_config,
                            metrics,
                            metadata,
                            artifact_url,
                            is_best,
                            created_at
                        )
                        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
                        """,
                        (
                            checkpoint.id,
                            checkpoint.workflow_id,
                            checkpoint.config_version,
                            json.dumps(checkpoint.runnable_config),
                            json.dumps(checkpoint.metrics),
                            json.dumps(checkpoint.metadata),
                            checkpoint.artifact_url,
                            1 if checkpoint.is_best else 0,
                            checkpoint.created_at.isoformat(),
                        ),
                    )
                    if is_best:
                        await conn.execute(
                            """
                            UPDATE agentensor_checkpoints
                               SET is_best = 0
                             WHERE workflow_id = ?
                               AND id != ?
                            """,
                            (workflow_id, checkpoint.id),
                        )
                    await conn.commit()
                except Exception:
                    await conn.rollback()
                    raise
                return checkpoint

    async def list_checkpoints(
        self,
        workflow_id: str,
        *,
        limit: int | None = None,
    ) -> list[AgentensorCheckpoint]:
        """Return persisted checkpoints ordered by config_version."""
        await self._ensure_initialized()
        query = (
            "SELECT id, workflow_id, config_version, runnable_config, metrics, "
            "metadata, artifact_url, is_best, created_at "
            "FROM agentensor_checkpoints WHERE workflow_id = ? "
            "ORDER BY config_version DESC"
        )
        params: list[object] = [workflow_id]
        if limit is not None:
            query += " LIMIT ?"
            params.append(limit)
        async with connect_sqlite(self._database_path) as conn:
            cursor = await conn.execute(query, tuple(params))
            rows = await cursor.fetchall()
        return [self._row_to_checkpoint(row) for row in rows]

    async def get_checkpoint(self, checkpoint_id: str) -> AgentensorCheckpoint:
        """Return a checkpoint by identifier or raise when missing."""
        await self._ensure_initialized()
        async with connect_sqlite(self._database_path) as conn:
            cursor = await conn.execute(
                """
                SELECT id, workflow_id, config_version, runnable_config, metrics,
                       metadata, artifact_url, is_best, created_at
                  FROM agentensor_checkpoints
                 WHERE id = ?
                """,
                (checkpoint_id,),
            )
            row = await cursor.fetchone()
        if row is None:
            msg = f"Checkpoint {checkpoint_id!r} not found."
            raise AgentensorCheckpointNotFoundError(msg)
        return self._row_to_checkpoint(row)

    async def latest_checkpoint(
        self,
        workflow_id: str,
    ) -> AgentensorCheckpoint | None:
        """Return the newest checkpoint if one exists for the workflow."""
        checkpoints = await self.list_checkpoints(workflow_id, limit=1)
        return checkpoints[0] if checkpoints else None

    async def _resolve_version(
        self,
        conn: aiosqlite.Connection,
        workflow_id: str,
        provided_version: int | None,
    ) -> int:
        """Resolve the next config version for the workflow."""
        if provided_version is not None:
            return provided_version
        cursor = await conn.execute(
            """
            SELECT COALESCE(MAX(config_version), 0) AS max_version
              FROM agentensor_checkpoints
             WHERE workflow_id = ?
            """,
            (workflow_id,),
        )
        row = await cursor.fetchone()
        max_version = row["max_version"] if row else 0
        return int(max_version) + 1

    async def _ensure_initialized(self) -> None:
        if self._initialized:
            return
        async with self._init_lock:
            if self._initialized:
                return
            await ensure_sqlite_schema(self._database_path)
            self._initialized = True

    @staticmethod
    def _row_to_checkpoint(row: aiosqlite.Row) -> AgentensorCheckpoint:
        return AgentensorCheckpoint(
            id=row["id"],
            workflow_id=row["workflow_id"],
            config_version=int(row["config_version"]),
            runnable_config=json.loads(row["runnable_config"]),
            metrics=json.loads(row["metrics"]),
            metadata=json.loads(row["metadata"]),
            artifact_url=row["artifact_url"],
            is_best=bool(row["is_best"]),
            created_at=datetime.fromisoformat(row["created_at"]),
        )


class PostgresAgentensorCheckpointStore(AgentensorCheckpointStore):
    """PostgreSQL-backed checkpoint store for production deployments."""

    def __init__(
        self,
        dsn: str,
        *,
        pool_min_size: int = 1,
        pool_max_size: int = 10,
        pool_timeout: float = 30.0,
        pool_max_idle: float = 300.0,
    ) -> None:
        """Initialize the PostgreSQL checkpoint store."""
        if _AsyncConnectionPool is None or _DictRowFactory is None:
            msg = "PostgreSQL backend requires psycopg[binary,pool] to be installed."
            raise RuntimeError(msg)

        self._dsn = dsn
        self._pool_min_size = pool_min_size
        self._pool_max_size = pool_max_size
        self._pool_timeout = pool_timeout
        self._pool_max_idle = pool_max_idle
        self._pool: Any | None = None
        self._lock = asyncio.Lock()
        self._init_lock = asyncio.Lock()
        self._initialized = False

    async def _get_pool(self) -> Any:
        """Return the connection pool, creating it if necessary."""
        if self._pool is not None:
            return self._pool

        async with self._init_lock:
            if self._pool is not None:
                return self._pool

            pool_class = _AsyncConnectionPool
            assert pool_class is not None  # mypy
            self._pool = pool_class(
                self._dsn,
                min_size=self._pool_min_size,
                max_size=self._pool_max_size,
                timeout=self._pool_timeout,
                max_idle=self._pool_max_idle,
                open=False,
                kwargs={
                    "autocommit": False,
                    "prepare_threshold": 0,
                    "row_factory": _DictRowFactory,
                },
            )
            await self._pool.open()
            return self._pool

    @asynccontextmanager
    async def _connection(self) -> AsyncIterator[Any]:
        pool = await self._get_pool()
        async with pool.connection() as conn:
            try:
                yield conn
                await conn.commit()
            except Exception:
                await conn.rollback()
                raise

    async def _ensure_initialized(self) -> None:
        if self._initialized:
            return

        async with self._init_lock:
            if self._initialized:
                return

            async with self._connection() as conn:
                for raw_stmt in POSTGRES_CHECKPOINT_MIGRATION.strip().split(";"):
                    stmt = raw_stmt.strip()
                    if stmt:
                        await conn.execute(stmt)

            self._initialized = True

    async def record_checkpoint(
        self,
        *,
        workflow_id: str,
        runnable_config: Mapping[str, object],
        metrics: Mapping[str, object],
        metadata: Mapping[str, object] | None = None,
        artifact_url: str | None = None,
        is_best: bool = False,
        config_version: int | None = None,
    ) -> AgentensorCheckpoint:
        """Persist a checkpoint and return the stored record."""
        await self._ensure_initialized()
        async with self._lock:
            async with self._connection() as conn:
                version = await self._resolve_version(conn, workflow_id, config_version)
                checkpoint = AgentensorCheckpoint(
                    workflow_id=workflow_id,
                    config_version=version,
                    runnable_config=dict(runnable_config),
                    metrics=dict(metrics),
                    metadata=dict(metadata or {}),
                    artifact_url=artifact_url,
                    is_best=is_best,
                )
                await conn.execute(
                    """
                    INSERT INTO agentensor_checkpoints (
                        id, workflow_id, config_version, runnable_config,
                        metrics, metadata, artifact_url, is_best, created_at
                    )
                    VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s)
                    """,
                    (
                        checkpoint.id,
                        checkpoint.workflow_id,
                        checkpoint.config_version,
                        json.dumps(checkpoint.runnable_config),
                        json.dumps(checkpoint.metrics),
                        json.dumps(checkpoint.metadata),
                        checkpoint.artifact_url,
                        checkpoint.is_best,
                        checkpoint.created_at,
                    ),
                )
                if is_best:
                    await conn.execute(
                        """
                        UPDATE agentensor_checkpoints
                           SET is_best = FALSE
                         WHERE workflow_id = %s
                           AND id != %s
                        """,
                        (workflow_id, checkpoint.id),
                    )
                return checkpoint

    async def list_checkpoints(
        self,
        workflow_id: str,
        *,
        limit: int | None = None,
    ) -> list[AgentensorCheckpoint]:
        """Return checkpoints for the workflow ordered newest-first."""
        await self._ensure_initialized()
        query = """
            SELECT id, workflow_id, config_version, runnable_config,
                   metrics, metadata, artifact_url, is_best, created_at
              FROM agentensor_checkpoints
             WHERE workflow_id = %s
             ORDER BY config_version DESC
        """
        params: list[object] = [workflow_id]
        if limit is not None:
            query += " LIMIT %s"
            params.append(limit)
        async with self._connection() as conn:
            cursor = await conn.execute(query, tuple(params))
            rows = await cursor.fetchall()
        return [self._row_to_checkpoint(row) for row in rows]

    async def get_checkpoint(self, checkpoint_id: str) -> AgentensorCheckpoint:
        """Return the checkpoint by identifier or raise when missing."""
        await self._ensure_initialized()
        async with self._connection() as conn:
            cursor = await conn.execute(
                """
                SELECT id, workflow_id, config_version, runnable_config,
                       metrics, metadata, artifact_url, is_best, created_at
                  FROM agentensor_checkpoints
                 WHERE id = %s
                """,
                (checkpoint_id,),
            )
            row = await cursor.fetchone()
        if row is None:
            msg = f"Checkpoint {checkpoint_id!r} not found."
            raise AgentensorCheckpointNotFoundError(msg)
        return self._row_to_checkpoint(row)

    async def latest_checkpoint(
        self,
        workflow_id: str,
    ) -> AgentensorCheckpoint | None:
        """Return the most recent checkpoint for the workflow if present."""
        checkpoints = await self.list_checkpoints(workflow_id, limit=1)
        return checkpoints[0] if checkpoints else None

    async def _resolve_version(
        self,
        conn: Any,
        workflow_id: str,
        provided_version: int | None,
    ) -> int:
        """Resolve the next config version for the workflow."""
        if provided_version is not None:
            return provided_version
        cursor = await conn.execute(
            """
            SELECT COALESCE(MAX(config_version), 0) AS max_version
              FROM agentensor_checkpoints
             WHERE workflow_id = %s
            """,
            (workflow_id,),
        )
        row = await cursor.fetchone()
        max_version = row["max_version"] if row else 0
        return int(max_version) + 1

    @staticmethod
    def _row_to_checkpoint(row: dict[str, Any]) -> AgentensorCheckpoint:
        """Convert a PostgreSQL row into an AgentensorCheckpoint instance."""
        created_at = row["created_at"]
        if isinstance(created_at, str):
            created_at = datetime.fromisoformat(created_at)

        runnable_config = row["runnable_config"]
        if isinstance(runnable_config, str):
            runnable_config = json.loads(runnable_config)

        metrics = row["metrics"]
        if isinstance(metrics, str):
            metrics = json.loads(metrics)

        metadata = row["metadata"]
        if isinstance(metadata, str):
            metadata = json.loads(metadata)

        return AgentensorCheckpoint(
            id=row["id"],
            workflow_id=row["workflow_id"],
            config_version=int(row["config_version"]),
            runnable_config=runnable_config,
            metrics=metrics,
            metadata=metadata,
            artifact_url=row["artifact_url"],
            is_best=bool(row["is_best"]),
            created_at=created_at,
        )

    async def close(self) -> None:
        """Close the connection pool."""
        if self._pool is not None:
            await self._pool.close()
            self._pool = None


__all__ = [
    "AgentensorCheckpointNotFoundError",
    "InMemoryAgentensorCheckpointStore",
    "POSTGRES_CHECKPOINT_MIGRATION",
    "PostgresAgentensorCheckpointStore",
    "SqliteAgentensorCheckpointStore",
]



"""Tests for workflow execution functions in tasks.py."""

from __future__ import annotations
from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch
from uuid import uuid4
import pytest
from pydantic import BaseModel


@pytest.fixture
def mock_run() -> MagicMock:
    """Create a mock pending run."""
    run = MagicMock()
    run.id = uuid4()
    run.workflow_version_id = uuid4()
    run.status = MagicMock()
    run.status.value = "pending"
    run.input_payload = {"test": "data"}
    run.runnable_config = None
    return run


@pytest.fixture
def mock_version() -> MagicMock:
    """Create a mock workflow version."""
    version = MagicMock()
    version.id = uuid4()
    version.workflow_id = uuid4()
    version.graph = {"nodes": [], "edges": []}
    version.runnable_config = None
    return version


class TestExtractOutput:
    """Tests for _extract_output function."""

    def test_extracts_dict_output(self) -> None:
        """Test that dict final state is wrapped correctly."""
        from orcheo_backend.worker.tasks import _extract_output

        final_state = {"key": "value", "count": 42}
        result = _extract_output(final_state)

        assert result == {"final_state": final_state}

    def test_extracts_pydantic_model_output(self) -> None:
        """Test that Pydantic model is dumped correctly."""
        from orcheo_backend.worker.tasks import _extract_output

        class SampleModel(BaseModel):
            name: str
            value: int

        model = SampleModel(name="test", value=123)
        result = _extract_output(model)

        assert result == {"final_state": {"name": "test", "value": 123}}

    def test_returns_none_for_unknown_type(self) -> None:
        """Test that unknown types return None."""
        from orcheo_backend.worker.tasks import _extract_output

        result = _extract_output("just a string")
        assert result is None

    def test_returns_none_for_none_input(self) -> None:
        """Test that None input returns None."""
        from orcheo_backend.worker.tasks import _extract_output

        result = _extract_output(None)
        assert result is None

    def test_extracts_object_with_model_dump(self) -> None:
        """Test extraction of non-Pydantic object with model_dump method."""
        from orcheo_backend.worker.tasks import _extract_output

        class CustomObject:
            def model_dump(self) -> dict[str, Any]:
                return {"custom": "data"}

        obj = CustomObject()
        result = _extract_output(obj)

        assert result == {"final_state": {"custom": "data"}}


class TestHandleExecutionFailure:
    """Tests for _handle_execution_failure function."""

    @pytest.mark.asyncio
    async def test_marks_run_as_failed(self, mock_run: MagicMock) -> None:
        """Test that run is marked as failed."""
        from orcheo_backend.worker.tasks import _handle_execution_failure

        mock_repo = AsyncMock()
        exception = ValueError("Workflow crashed")

        with patch(
            "orcheo_backend.app.dependencies.get_repository", return_value=mock_repo
        ):
            with patch("orcheo_backend.worker.tasks.logger"):
                result = await _handle_execution_failure(mock_run, exception)

        assert result["status"] == "failed"
        assert "Workflow crashed" in result["error"]
        mock_repo.mark_run_failed.assert_called_once_with(
            mock_run.id, actor="worker", error="Workflow crashed"
        )

    @pytest.mark.asyncio
    async def test_logs_exception_when_mark_failed_fails(
        self, mock_run: MagicMock
    ) -> None:
        """Test that exception is logged if mark_run_failed fails."""
        from orcheo_backend.worker.tasks import _handle_execution_failure

        mock_repo = AsyncMock()
        mock_repo.mark_run_failed = AsyncMock(
            side_effect=RuntimeError("Database error")
        )
        exception = ValueError("Original error")

        with patch(
            "orcheo_backend.app.dependencies.get_repository", return_value=mock_repo
        ):
            with patch("orcheo_backend.worker.tasks.logger") as mock_logger:
                result = await _handle_execution_failure(mock_run, exception)

        # Should still return the original error
        assert result["status"] == "failed"
        assert "Original error" in result["error"]
        # Should log both the original exception and the mark_run_failed failure
        assert mock_logger.exception.call_count == 2

    @pytest.mark.asyncio
    async def test_returns_error_message_from_exception(
        self, mock_run: MagicMock
    ) -> None:
        """Test that error message comes from the exception."""
        from orcheo_backend.worker.tasks import _handle_execution_failure

        mock_repo = AsyncMock()
        exception = RuntimeError("Specific runtime error")

        with patch(
            "orcheo_backend.app.dependencies.get_repository", return_value=mock_repo
        ):
            with patch("orcheo_backend.worker.tasks.logger"):
                result = await _handle_execution_failure(mock_run, exception)

        assert result["error"] == "Specific runtime error"

    @pytest.mark.asyncio
    async def test_logs_exception_when_history_mark_failed_fails(
        self, mock_run: MagicMock
    ) -> None:
        """Test history failures are logged without changing the return payload."""
        from orcheo_backend.worker.tasks import _handle_execution_failure

        mock_repo = AsyncMock()
        mock_history = AsyncMock()
        mock_history.mark_failed = AsyncMock(side_effect=RuntimeError("history down"))
        exception = RuntimeError("Execution exploded")

        with patch(
            "orcheo_backend.app.dependencies.get_repository", return_value=mock_repo
        ):
            with patch("orcheo_backend.worker.tasks.logger") as mock_logger:
                result = await _handle_execution_failure(
                    mock_run, exception, history_store=mock_history
                )

        assert result == {"status": "failed", "error": "Execution exploded"}
        mock_history.append_step.assert_awaited_once_with(
            str(mock_run.id), {"status": "error", "error": "Execution exploded"}
        )
        mock_history.mark_failed.assert_awaited_once_with(
            str(mock_run.id), "Execution exploded"
        )
        assert mock_logger.exception.call_count == 2


class TestExecuteWorkflow:
    """Tests for _execute_workflow function."""

    @pytest.mark.asyncio
    async def test_successful_execution(
        self, mock_run: MagicMock, mock_version: MagicMock
    ) -> None:
        """Test successful workflow execution."""
        from orcheo_backend.worker.tasks import _execute_workflow

        mock_repo = AsyncMock()
        mock_repo.get_version = AsyncMock(return_value=mock_version)
        mock_repo.mark_run_succeeded = AsyncMock()
        mock_history = AsyncMock()

        async def _step_stream() -> Any:
            yield {"node_a": {"status": "running"}}

        mock_graph = MagicMock()
        mock_compiled = MagicMock()
        mock_compiled.astream = MagicMock(return_value=_step_stream())
        mock_compiled.aget_state = AsyncMock(return_value=MagicMock(values={}))
        mock_graph.compile = MagicMock(return_value=mock_compiled)

        mock_checkpointer = MagicMock()
        mock_checkpointer.__aenter__ = AsyncMock(return_value=mock_checkpointer)
        mock_checkpointer.__aexit__ = AsyncMock(return_value=None)

        with patch(
            "orcheo_backend.app.dependencies.get_repository", return_value=mock_repo
        ):
            with patch(
                "orcheo_backend.app.dependencies.get_vault",
                return_value=MagicMock(),
            ):
                with patch(
                    "orcheo_backend.app.dependencies.get_history_store",
                    return_value=mock_history,
                ):
                    with patch(
                        "orcheo.graph.builder.build_graph", return_value=mock_graph
                    ):
                        with patch(
                            "orcheo.persistence.create_checkpointer",
                            return_value=mock_checkpointer,
                        ):
                            with patch(
                                "orcheo.persistence.create_graph_store",
                                return_value=mock_checkpointer,
                            ):
                                with patch(
                                    "orcheo_backend.app.workflow_execution._build_initial_state",
                                    return_value={},
                                ):
                                    with patch("orcheo.config.get_settings"):
                                        with patch(
                                            "orcheo.runtime.runnable_config.merge_runnable_configs"
                                        ) as mock_merge:
                                            mock_config = MagicMock()
                                            mock_config.to_runnable_config = MagicMock(
                                                return_value={}
                                            )
                                            mock_config.to_state_config = MagicMock(
                                                return_value={}
                                            )
                                            mock_config.to_json_config = MagicMock(
                                                return_value={}
                                            )
                                            mock_config.tags = []
                                            mock_config.callbacks = []
                                            mock_config.metadata = {}
                                            mock_config.run_name = None
                                            mock_merge.return_value = mock_config

                                            with patch(
                                                "orcheo.runtime.credentials.credential_resolution"
                                            ):
                                                result = await _execute_workflow(
                                                    mock_run
                                                )

        assert result["status"] == "succeeded"
        mock_repo.mark_run_succeeded.assert_called_once()
        mock_graph.compile.assert_called_once_with(
            checkpointer=mock_checkpointer,
            store=mock_checkpointer,
        )
        mock_history.start_run.assert_awaited_once()
        mock_history.append_step.assert_awaited()
        mock_history.mark_completed.assert_awaited_once()

    @pytest.mark.asyncio
    async def test_handles_execution_exception(
        self, mock_run: MagicMock, mock_version: MagicMock
    ) -> None:
        """Test that execution exceptions are handled."""
        from orcheo_backend.worker.tasks import _execute_workflow

        mock_repo = AsyncMock()
        mock_repo.get_version = AsyncMock(side_effect=RuntimeError("Version not found"))
        mock_repo.mark_run_failed = AsyncMock()
        mock_history = AsyncMock()

        with patch(
            "orcheo_backend.app.dependencies.get_repository", return_value=mock_repo
        ):
            with patch(
                "orcheo_backend.app.dependencies.get_vault",
                return_value=MagicMock(),
            ):
                with patch(
                    "orcheo_backend.app.dependencies.get_history_store",
                    return_value=mock_history,
                ):
                    with patch("orcheo.config.get_settings"):
                        with patch("orcheo_backend.worker.tasks.logger"):
                            result = await _execute_workflow(mock_run)

        assert result["status"] == "failed"
        assert "Version not found" in result["error"]
        mock_history.append_step.assert_awaited_once()
        mock_history.mark_failed.assert_awaited_once()


class TestHistoryHelpers:
    """Tests for helper functions that persist run history."""

    @pytest.mark.asyncio
    async def test_start_history_record_logs_history_error(self) -> None:
        """Test _start_history_record catches history store errors."""
        from orcheo_backend.worker.tasks import _start_history_record

        history_error_cls = type("HistoryError", (Exception,), {})
        mock_history = AsyncMock()
        mock_history.start_run = AsyncMock(side_effect=history_error_cls("oops"))
        mock_config = MagicMock()
        mock_config.to_json_config.return_value = {"x": 1}
        mock_config.tags = []
        mock_config.callbacks = []
        mock_config.metadata = {}
        mock_config.run_name = None

        with patch("orcheo_backend.worker.tasks.logger") as mock_logger:
            await _start_history_record(
                history_store=mock_history,
                workflow_id=str(uuid4()),
                execution_id=str(uuid4()),
                inputs={"k": "v"},
                merged_config=mock_config,
                history_error_cls=history_error_cls,
            )

        mock_logger.exception.assert_called_once()

    @pytest.mark.asyncio
    async def test_stream_run_history_steps_logs_history_error(self) -> None:
        """Test _stream_run_history_steps catches append failures."""
        from orcheo_backend.worker.tasks import _stream_run_history_steps

        history_error_cls = type("HistoryError", (Exception,), {})
        mock_history = AsyncMock()
        mock_history.append_step = AsyncMock(side_effect=history_error_cls("append"))
        mock_compiled = MagicMock()

        async def _step_stream() -> Any:
            yield {"node": {"status": "running"}}

        mock_compiled.astream = MagicMock(return_value=_step_stream())

        with patch("orcheo_backend.worker.tasks.logger") as mock_logger:
            await _stream_run_history_steps(
                compiled=mock_compiled,
                state={},
                runtime_config={},
                history_store=mock_history,
                execution_id=str(uuid4()),
                history_error_cls=history_error_cls,
            )

        mock_logger.exception.assert_called_once()

    @pytest.mark.asyncio
    async def test_mark_history_completed_logs_history_error(self) -> None:
        """Test _mark_history_completed catches completion failures."""
        from orcheo_backend.worker.tasks import _mark_history_completed

        history_error_cls = type("HistoryError", (Exception,), {})
        mock_history = AsyncMock()
        mock_history.mark_completed = AsyncMock(side_effect=history_error_cls("done"))

        with patch("orcheo_backend.worker.tasks.logger") as mock_logger:
            await _mark_history_completed(
                history_store=mock_history,
                execution_id=str(uuid4()),
                history_error_cls=history_error_cls,
            )

        mock_history.append_step.assert_awaited_once()
        mock_logger.exception.assert_called_once()


class TestExecuteRunAsync:
    """Tests for _execute_run_async function."""

    @pytest.mark.asyncio
    async def test_returns_error_when_load_fails(self) -> None:
        """Test that load failure returns error."""
        from orcheo_backend.worker.tasks import _execute_run_async

        with patch(
            "orcheo_backend.worker.tasks._load_and_validate_run",
            return_value=(None, {"status": "failed", "error": "Not found"}),
        ):
            result = await _execute_run_async(str(uuid4()))

        assert result["status"] == "failed"

    @pytest.mark.asyncio
    async def test_returns_error_when_mark_started_fails(
        self, mock_run: MagicMock
    ) -> None:
        """Test that mark_started failure returns error."""
        from orcheo_backend.worker.tasks import _execute_run_async

        with patch(
            "orcheo_backend.worker.tasks._load_and_validate_run",
            return_value=(mock_run, None),
        ):
            with patch(
                "orcheo_backend.worker.tasks._mark_run_started",
                return_value={"status": "skipped", "reason": "Already started"},
            ):
                result = await _execute_run_async(str(mock_run.id))

        assert result["status"] == "skipped"

    @pytest.mark.asyncio
    async def test_calls_execute_workflow_on_success(self, mock_run: MagicMock) -> None:
        """Test that _execute_workflow is called when validation succeeds."""
        from orcheo_backend.worker.tasks import _execute_run_async

        with patch(
            "orcheo_backend.worker.tasks._load_and_validate_run",
            return_value=(mock_run, None),
        ):
            with patch(
                "orcheo_backend.worker.tasks._mark_run_started",
                return_value=None,
            ):
                with patch(
                    "orcheo_backend.worker.tasks._execute_workflow",
                    return_value={"status": "succeeded"},
                ) as mock_execute:
                    result = await _execute_run_async(str(mock_run.id))

        assert result["status"] == "succeeded"
        mock_execute.assert_called_once_with(mock_run)


class TestDispatchCronTriggersAsync:
    """Tests for _dispatch_cron_triggers_async function."""

    @pytest.mark.asyncio
    async def test_dispatches_due_runs(self) -> None:
        """Test that due cron runs are dispatched."""
        from orcheo_backend.worker.tasks import _dispatch_cron_triggers_async

        mock_run1 = MagicMock()
        mock_run1.id = uuid4()
        mock_run2 = MagicMock()
        mock_run2.id = uuid4()

        mock_repo = AsyncMock()
        mock_repo.dispatch_due_cron_runs = AsyncMock(
            return_value=[mock_run1, mock_run2]
        )

        with patch(
            "orcheo_backend.app.dependencies.get_repository", return_value=mock_repo
        ):
            result = await _dispatch_cron_triggers_async()

        assert len(result) == 2
        assert str(mock_run1.id) in result
        assert str(mock_run2.id) in result

    @pytest.mark.asyncio
    async def test_returns_empty_list_when_no_runs_due(self) -> None:
        """Test that empty list is returned when no runs are due."""
        from orcheo_backend.worker.tasks import _dispatch_cron_triggers_async

        mock_repo = AsyncMock()
        mock_repo.dispatch_due_cron_runs = AsyncMock(return_value=[])

        with patch(
            "orcheo_backend.app.dependencies.get_repository", return_value=mock_repo
        ):
            result = await _dispatch_cron_triggers_async()

        assert result == []

    @pytest.mark.asyncio
    async def test_returns_run_ids(self) -> None:
        """Test that dispatched runs are returned."""
        from orcheo_backend.worker.tasks import _dispatch_cron_triggers_async

        mock_run = MagicMock()
        mock_run.id = uuid4()

        mock_repo = AsyncMock()
        mock_repo.dispatch_due_cron_runs = AsyncMock(return_value=[mock_run])

        with patch(
            "orcheo_backend.app.dependencies.get_repository", return_value=mock_repo
        ):
            result = await _dispatch_cron_triggers_async()

        assert result == [str(mock_run.id)]



"""Runtime loader for enabled Orcheo plugins."""

from __future__ import annotations
import importlib.metadata
import logging
import sys
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from orcheo.edges.registry import edge_registry
from orcheo.listeners.registry import listener_registry
from orcheo.nodes.agent_tools.registry import tool_registry
from orcheo.nodes.registry import registry
from orcheo.plugins.api import PluginAPI, PluginRegistrations
from orcheo.plugins.compatibility import check_manifest_compatibility
from orcheo.plugins.manager import (
    PLUGIN_ENTRYPOINT_GROUP,
    _distribution_to_manifest,
    _site_packages,
)
from orcheo.plugins.models import PluginManifest
from orcheo.plugins.paths import build_storage_paths
from orcheo.plugins.state import load_desired_state, load_lock_state
from orcheo.triggers.registry import trigger_registry


logger = logging.getLogger(__name__)


@dataclass(slots=True)
class PluginLoadResult:
    """Summary for one attempted plugin load."""

    name: str
    loaded: bool
    manifest: PluginManifest | None = None
    error: str | None = None


@dataclass(slots=True)
class PluginLoadReport:
    """Aggregate report for one plugin load pass."""

    generation: int
    results: list[PluginLoadResult]


_LOADED = False
_GENERATION = 0
_REPORT = PluginLoadReport(generation=0, results=[])
_REGISTERED = PluginRegistrations()


def _purge_module_roots(module_roots: Iterable[str]) -> None:
    """Remove loaded plugin modules so a reload can import fresh code."""
    roots = {root for root in module_roots if root}
    if not roots:
        return
    for module_name in sorted(sys.modules, reverse=True):
        if any(
            module_name == root or module_name.startswith(f"{root}.") for root in roots
        ):
            sys.modules.pop(module_name, None)


def _clear_registered_components() -> None:
    """Remove plugin-provided components from process-local registries."""
    global _REGISTERED  # noqa: PLW0603
    for name in reversed(_REGISTERED.nodes):
        registry.unregister(name)
    for name in reversed(_REGISTERED.edges):
        edge_registry.unregister(name)
    for name in reversed(_REGISTERED.agent_tools):
        tool_registry.unregister(name)
    for name in reversed(_REGISTERED.listeners):
        listener_registry.unregister(name)
    for name in reversed(_REGISTERED.triggers):
        trigger_registry.unregister(name)
    _purge_module_roots(_REGISTERED.module_roots)
    _REGISTERED = PluginRegistrations()


def _rollback_plugin_registrations(registrations: PluginRegistrations) -> None:
    """Roll back a plugin's partial registrations after a load failure."""
    for name in reversed(registrations.nodes):
        registry.unregister(name)
    for name in reversed(registrations.edges):
        edge_registry.unregister(name)
    for name in reversed(registrations.agent_tools):
        tool_registry.unregister(name)
    for name in reversed(registrations.listeners):
        listener_registry.unregister(name)
    for name in reversed(registrations.triggers):
        trigger_registry.unregister(name)
    _purge_module_roots(registrations.module_roots)


def _plugin_site_packages() -> Path:
    paths = build_storage_paths()
    return _site_packages(Path(paths.install_dir))


def _enabled_locked_plugin_names() -> set[str]:
    paths = build_storage_paths()
    desired = {
        record.name: record for record in load_desired_state(Path(paths.state_file))
    }
    locked = {record.name: record for record in load_lock_state(Path(paths.lock_file))}
    return {
        name for name, record in desired.items() if record.enabled and name in locked
    }


def _iter_plugin_distributions(
    site_packages: Path,
    *,
    names: Iterable[str],
) -> list[importlib.metadata.Distribution]:
    wanted = {name.lower() for name in names}
    matches: list[importlib.metadata.Distribution] = []
    for distribution in importlib.metadata.distributions(path=[str(site_packages)]):
        metadata_name = str(distribution.metadata.get("Name", ""))
        if metadata_name.lower() not in wanted:
            continue
        if any(
            entry.group == PLUGIN_ENTRYPOINT_GROUP
            for entry in distribution.entry_points
        ):
            matches.append(distribution)
    return matches


def _ensure_sys_path(site_packages: Path) -> None:
    if site_packages.exists() and str(site_packages) not in sys.path:
        sys.path.insert(0, str(site_packages))


def _entry_point_module_root(entry_point: importlib.metadata.EntryPoint) -> str:
    """Return the top-level module path for one plugin entry point."""
    module_name = getattr(entry_point, "module", "")
    if module_name:
        return str(module_name)
    return str(entry_point.value).split(":", maxsplit=1)[0].strip()


def load_enabled_plugins(  # noqa: C901, PLR0915
    *, force: bool = False
) -> PluginLoadReport:
    """Load all enabled plugins from the managed plugin environment."""
    global _GENERATION, _LOADED, _REPORT  # noqa: PLW0603
    if _LOADED and not force:
        return _REPORT
    _clear_registered_components()

    site_packages = _plugin_site_packages()
    enabled_names = _enabled_locked_plugin_names()
    if not enabled_names or not site_packages.exists():
        _GENERATION += 1
        _LOADED = True
        _REPORT = PluginLoadReport(generation=_GENERATION, results=[])
        return _REPORT

    _ensure_sys_path(site_packages)
    results: list[PluginLoadResult] = []
    distributions = _iter_plugin_distributions(site_packages, names=enabled_names)
    discovered_names = {
        str(distribution.metadata.get("Name", "")) for distribution in distributions
    }
    for missing_name in sorted(
        enabled_names - {name.lower() for name in discovered_names}
    ):
        results.append(
            PluginLoadResult(
                name=missing_name,
                loaded=False,
                error=(
                    "Plugin is enabled and locked but not installed in the "
                    "plugin environment."
                ),
            )
        )

    for distribution in distributions:
        api = PluginAPI()
        manifest, _manifest_hash = _distribution_to_manifest(distribution)
        issues = check_manifest_compatibility(manifest)
        if issues:
            error = "; ".join(issues)
            logger.warning("Skipping incompatible plugin %s: %s", manifest.name, error)
            results.append(
                PluginLoadResult(
                    name=manifest.name,
                    loaded=False,
                    manifest=manifest,
                    error=error,
                )
            )
            continue
        entry_points = [
            entry
            for entry in distribution.entry_points
            if entry.group == PLUGIN_ENTRYPOINT_GROUP
        ]
        if not entry_points:
            results.append(
                PluginLoadResult(
                    name=manifest.name,
                    loaded=False,
                    manifest=manifest,
                    error="No plugin entry points registered.",
                )
            )
            continue
        try:
            for entry_point in entry_points:
                plugin = entry_point.load()
                register = getattr(plugin, "register", None)
                if not callable(register):
                    msg = (
                        f"Plugin entry point '{entry_point.name}' has no "
                        "register(api) hook."
                    )
                    raise TypeError(msg)
                register(api)
                module_root = _entry_point_module_root(entry_point)
                if module_root and module_root not in api.registrations.module_roots:
                    api.registrations.module_roots.append(module_root)
        except Exception as exc:  # pragma: no cover - covered in loader tests
            _rollback_plugin_registrations(api.registrations)
            logger.exception("Failed to load plugin %s", manifest.name)
            results.append(
                PluginLoadResult(
                    name=manifest.name,
                    loaded=False,
                    manifest=manifest,
                    error=str(exc),
                )
            )
            continue
        results.append(
            PluginLoadResult(
                name=manifest.name,
                loaded=True,
                manifest=manifest,
            )
        )
        _REGISTERED.nodes.extend(api.registrations.nodes)
        _REGISTERED.edges.extend(api.registrations.edges)
        _REGISTERED.agent_tools.extend(api.registrations.agent_tools)
        _REGISTERED.listeners.extend(api.registrations.listeners)
        _REGISTERED.triggers.extend(api.registrations.triggers)
        _REGISTERED.module_roots.extend(api.registrations.module_roots)

    _GENERATION += 1
    _LOADED = True
    _REPORT = PluginLoadReport(generation=_GENERATION, results=results)
    return _REPORT


def ensure_plugins_loaded() -> PluginLoadReport:
    """Load enabled plugins once per process."""
    return load_enabled_plugins(force=False)


def invalidate_plugin_loader() -> None:
    """Invalidate process-local plugin load state after lifecycle changes."""
    global _LOADED, _REPORT  # noqa: PLW0603
    _LOADED = False
    _REPORT = PluginLoadReport(generation=_GENERATION, results=[])
    _clear_registered_components()


def reset_plugin_loader_for_tests() -> None:
    """Reset the process-local loader state for isolated tests."""
    invalidate_plugin_loader()


__all__ = [
    "PluginLoadReport",
    "PluginLoadResult",
    "ensure_plugins_loaded",
    "invalidate_plugin_loader",
    "load_enabled_plugins",
    "reset_plugin_loader_for_tests",
]



"""Tests for AgentReplyExtractorNode (lines 417-429 of ai.py)."""

from __future__ import annotations
import pytest
from langchain_core.messages import AIMessage, HumanMessage
from langchain_core.runnables import RunnableConfig
from orcheo.graph.state import State
from orcheo.nodes.ai import AgentReplyExtractorNode


@pytest.mark.asyncio
async def test_extractor_returns_dict_assistant_content() -> None:
    """Dict-style assistant message: returns its content string (lines 419-423)."""
    node = AgentReplyExtractorNode(name="extractor")
    state: State = {
        "messages": [
            {"role": "user", "content": "Hello"},
            {"role": "assistant", "content": "Hi there!"},
        ],
        "inputs": {},
        "results": {},
        "structured_response": None,
    }
    result = await node.run(state, RunnableConfig())
    assert result == {"agent_reply": "Hi there!"}


@pytest.mark.asyncio
async def test_extractor_skips_dict_assistant_with_empty_content() -> None:
    """Dict assistant message with empty content is skipped; fallback is
    returned (line 422 false branch).
    """
    node = AgentReplyExtractorNode(name="extractor")
    state: State = {
        "messages": [{"role": "assistant", "content": ""}],
        "inputs": {},
        "results": {},
        "structured_response": None,
    }
    result = await node.run(state, RunnableConfig())
    assert result == {"agent_reply": node.fallback_message}


@pytest.mark.asyncio
async def test_extractor_skips_dict_non_assistant_role() -> None:
    """Dict message with non-assistant role is not extracted (line 420 false branch)."""
    node = AgentReplyExtractorNode(name="extractor")
    state: State = {
        "messages": [{"role": "user", "content": "Question"}],
        "inputs": {},
        "results": {},
        "structured_response": None,
    }
    result = await node.run(state, RunnableConfig())
    assert result == {"agent_reply": node.fallback_message}


@pytest.mark.asyncio
async def test_extractor_returns_ai_message_string_content() -> None:
    """AIMessage with string content is returned directly (lines 424-428,
    isinstance str True).
    """
    node = AgentReplyExtractorNode(name="extractor")
    state: State = {
        "messages": [HumanMessage(content="Q"), AIMessage(content="Answer")],
        "inputs": {},
        "results": {},
        "structured_response": None,
    }
    result = await node.run(state, RunnableConfig())
    assert result == {"agent_reply": "Answer"}


@pytest.mark.asyncio
async def test_extractor_stringifies_ai_message_non_string_content() -> None:
    """AIMessage with non-string content is converted via str() (line 427
    str() branch).
    """
    node = AgentReplyExtractorNode(name="extractor")
    content = [{"type": "text", "text": "structured"}]
    state: State = {
        "messages": [AIMessage(content=content)],  # type: ignore[list-item]
        "inputs": {},
        "results": {},
        "structured_response": None,
    }
    result = await node.run(state, RunnableConfig())
    assert result == {"agent_reply": str(content)}


@pytest.mark.asyncio
async def test_extractor_returns_fallback_when_no_assistant_message() -> None:
    """Fallback message returned when messages list has no assistant turn (line 429)."""
    node = AgentReplyExtractorNode(
        name="extractor", fallback_message="No reply available"
    )
    state: State = {
        "messages": [HumanMessage(content="Question")],
        "inputs": {},
        "results": {},
        "structured_response": None,
    }
    result = await node.run(state, RunnableConfig())
    assert result == {"agent_reply": "No reply available"}


@pytest.mark.asyncio
async def test_extractor_returns_fallback_for_empty_messages() -> None:
    """Fallback returned when messages list is empty (line 417-418, 429)."""
    node = AgentReplyExtractorNode(name="extractor")
    state: State = {
        "messages": [],
        "inputs": {},
        "results": {},
        "structured_response": None,
    }
    result = await node.run(state, RunnableConfig())
    assert result["agent_reply"] == node.fallback_message



from __future__ import annotations
from uuid import UUID
from fastapi.testclient import TestClient
from orcheo.models import CredentialScope
from orcheo.vault import InMemoryCredentialVault


def _create_workflow(api_client: TestClient) -> str:
    response = api_client.post(
        "/api/workflows",
        json={"name": "Node Flow", "actor": "tester"},
    )
    assert response.status_code == 201
    return response.json()["id"]


def test_node_execution_with_set_variable_node(api_client: TestClient) -> None:
    """Test executing a SetVariableNode in isolation."""
    response = api_client.post(
        "/api/nodes/execute",
        json={
            "node_config": {
                "type": "SetVariableNode",
                "name": "test_node",
                "variables": {"foo": "bar", "count": 42},
            },
            "inputs": {},
        },
    )

    assert response.status_code == 200
    result = response.json()
    assert result["status"] == "success"
    assert result["result"] == {"foo": "bar", "count": 42}
    assert result["error"] is None


def test_node_execution_with_delay_node(api_client: TestClient) -> None:
    """Test executing a DelayNode in isolation."""
    response = api_client.post(
        "/api/nodes/execute",
        json={
            "node_config": {
                "type": "DelayNode",
                "name": "delay_test",
                "duration_seconds": 0.01,
            },
            "inputs": {},
        },
    )

    assert response.status_code == 200
    result = response.json()
    assert result["status"] == "success"
    assert result["result"] == {"duration_seconds": 0.01}
    assert result["error"] is None


def test_node_execution_resolves_credentials(api_client: TestClient) -> None:
    """Placeholders in node config should resolve through the vault."""

    workflow_id = UUID(_create_workflow(api_client))
    vault: InMemoryCredentialVault = api_client.app.state.vault  # type: ignore[attr-defined]
    vault.create_credential(
        name="telegram_token",
        provider="telegram",
        scopes=["bot"],
        secret="resolved-token",
        actor="tester",
        scope=CredentialScope.for_workflows(workflow_id),
    )

    response = api_client.post(
        "/api/nodes/execute",
        json={
            "node_config": {
                "type": "SetVariableNode",
                "name": "store_secret",
                "variables": {"token": "[[telegram_token]]"},
            },
            "inputs": {},
            "workflow_id": str(workflow_id),
        },
    )

    assert response.status_code == 200
    result = response.json()
    assert result["status"] == "success"
    assert result["result"] == {"token": "resolved-token"}


def test_node_execution_with_inputs(api_client: TestClient) -> None:
    """Test executing a node with custom inputs."""
    response = api_client.post(
        "/api/nodes/execute",
        json={
            "node_config": {
                "type": "SetVariableNode",
                "name": "input_test",
                "variables": {"output": "processed"},
            },
            "inputs": {"input_value": "test_data"},
        },
    )

    assert response.status_code == 200
    result = response.json()
    assert result["status"] == "success"
    assert result["result"]["output"] == "processed"


def test_node_execution_missing_type(api_client: TestClient) -> None:
    """Test that missing node type returns 400 error."""
    response = api_client.post(
        "/api/nodes/execute",
        json={
            "node_config": {
                "name": "missing_type",
            },
            "inputs": {},
        },
    )

    assert response.status_code == 400
    result = response.json()
    assert "type" in result["detail"]


def test_node_execution_unknown_type(api_client: TestClient) -> None:
    """Test that unknown node type returns 400 error."""
    response = api_client.post(
        "/api/nodes/execute",
        json={
            "node_config": {
                "type": "NonExistentNode",
                "name": "unknown",
            },
            "inputs": {},
        },
    )

    assert response.status_code == 400
    result = response.json()
    assert "Unknown node type" in result["detail"]


def test_node_execution_invalid_config(api_client: TestClient) -> None:
    """Test that invalid node configuration returns error status."""
    response = api_client.post(
        "/api/nodes/execute",
        json={
            "node_config": {
                "type": "DelayNode",
                # Missing required 'name' field
                "duration_seconds": 1,
            },
            "inputs": {},
        },
    )

    assert response.status_code == 200
    result = response.json()
    assert result["status"] == "error"
    assert result["error"] is not None
    assert result["result"] is None


def test_node_execution_with_workflow_context(api_client: TestClient) -> None:
    """Test executing a node with workflow_id for credential context."""
    workflow_id = _create_workflow(api_client)

    response = api_client.post(
        "/api/nodes/execute",
        json={
            "node_config": {
                "type": "SetVariableNode",
                "name": "context_test",
                "variables": {"status": "executed"},
            },
            "inputs": {},
            "workflow_id": workflow_id,
        },
    )

    assert response.status_code == 200
    result = response.json()
    assert result["status"] == "success"
    assert result["result"]["status"] == "executed"



"""Show workflow command."""

from __future__ import annotations
from collections.abc import Mapping
from typing import Annotated, Any
import typer
from rich.markup import escape
from orcheo_sdk.cli.output import format_datetime, print_json, render_json, render_table
from orcheo_sdk.cli.utils import load_with_cache
from orcheo_sdk.cli.workflow.app import WorkflowIdArgument, _state, workflow_app
from orcheo_sdk.cli.workflow.inputs import _cache_notice
from orcheo_sdk.cli.workflow.mermaid import _mermaid_from_graph
from orcheo_sdk.cli.workflow.reminders import (
    attach_workflow_vault_reminder,
    describe_workflow_vault_reminder,
    fetch_workflow_vault_readiness,
)
from orcheo_sdk.services import show_workflow_data


def _print_workflow_vault_reminder(
    console: Any,
    readiness: dict[str, object] | None,
    *,
    leading_newline: bool = False,
) -> None:
    prefix = "\n" if leading_newline else ""
    reminder = describe_workflow_vault_reminder(readiness)
    if reminder is None:
        return
    console.print(f"{prefix}[dim]Vault reminder: {escape(reminder)}[/dim]")


@workflow_app.command("show")
def show_workflow(
    ctx: typer.Context,
    workflow_id: WorkflowIdArgument,
    version: Annotated[
        int | None,
        typer.Option("--version", "-v", help="Show a specific version number"),
    ] = None,
) -> None:
    """Display details about a workflow, including its latest version and runs."""
    state = _state(ctx)
    workflow, workflow_cached, workflow_stale = load_with_cache(
        state,
        f"workflow:{workflow_id}",
        lambda: state.client.get(f"/api/workflows/{workflow_id}"),
    )
    if workflow_cached:
        _cache_notice(state, f"workflow {workflow_id}", workflow_stale)

    # Only fetch versions list when no specific version is requested
    versions = None
    if version is None:
        versions, _, _ = load_with_cache(
            state,
            f"workflow:{workflow_id}:versions",
            lambda: state.client.get(f"/api/workflows/{workflow_id}/versions"),
        )

    runs, runs_cached, runs_stale = load_with_cache(
        state,
        f"workflow:{workflow_id}:runs",
        lambda: state.client.get(f"/api/workflows/{workflow_id}/runs"),
    )
    if runs_cached:
        _cache_notice(state, f"workflow {workflow_id} runs", runs_stale)

    data = show_workflow_data(
        state.client,
        workflow_id,
        workflow=workflow,
        versions=versions,
        runs=runs,
        target_version=version,
    )
    readiness = fetch_workflow_vault_readiness(state.client, workflow_id)

    if not state.human:
        print_json(attach_workflow_vault_reminder(data, readiness))
        return

    workflow_details = data["workflow"]
    selected_version = data.get("selected_version")
    recent_runs = data.get("recent_runs", [])

    state.console.print("[bold]Publish status[/bold]")
    state.console.print(
        f"Status: {'Public' if workflow_details.get('is_public') else 'Private'}"
    )
    state.console.print(
        f"Require login: {'Yes' if workflow_details.get('require_login') else 'No'}"
    )
    published_at = workflow_details.get("published_at")
    state.console.print(
        f"Published at: {format_datetime(published_at) if published_at else '-'}"
    )
    share_url = workflow_details.get("share_url")
    state.console.print(f"Share URL: {share_url or '-'}\n")

    render_json(state.console, workflow_details, title="Workflow")

    if selected_version:
        graph_raw = selected_version.get("graph", {})
        graph = graph_raw if isinstance(graph_raw, Mapping) else {}
        mermaid = _mermaid_from_graph(graph)
        version_num = selected_version.get("version")
        version_label = f"Version {version_num}" if version else "Latest version"
        state.console.print(f"\n[bold]{version_label}[/bold]")
        render_json(state.console, selected_version)
        state.console.print("\n[bold]Mermaid[/bold]")
        state.console.print(mermaid)

    if recent_runs:
        rows = [
            [
                item.get("id"),
                item.get("status"),
                item.get("triggered_by"),
                item.get("created_at"),
            ]
            for item in recent_runs
        ]
        render_table(
            state.console,
            title="Recent runs",
            columns=["ID", "Status", "Actor", "Created at"],
            rows=rows,
            column_overflow={"ID": "fold"},
        )

    _print_workflow_vault_reminder(
        state.console,
        readiness,
        leading_newline=True,
    )


__all__ = ["show_workflow"]
